@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -32,7 +32,6 @@ class Serial(base.Layer):
         tl.Dense(128),
         tl.Relu(),
         tl.Dense(10),
-        tl.LogSoftmax()
       )
 
   A Serial combinator uses stack semantics to manage data for its sublayers.
@@ -67,11 +66,11 @@ class Serial(base.Layer):
   def forward(self, xs):
     """Executes this layer as part of a forward pass through the model."""
     self._validate_forward_inputs(xs)
-    state, weights = self.state, self.weights
-    rngs = _split_rngs(self.rng, self._n_layers)
-    if not self.sublayers:  # No-op: leave args unchanged.
+    if not self.sublayers:  # No-op: outputs = inputs
       return xs
 
+    state, weights = self.state, self.weights
+    rngs = _split_rngs(self.rng, self._n_layers)
     stack = xs
     new_state = []
     n_layers = self._n_layers
@@ -89,7 +88,7 @@ class Serial(base.Layer):
       outputs, s = layer.pure_fn(inputs, w, s, rng, use_cache=True)
       stack = outputs_onto_stack(outputs, stack, layer.n_in)
       new_state.append(s)
-    self.state = new_state
+    self.state = tuple(new_state)
     return stack
 
   # pylint: disable=protected-access
@@ -109,8 +108,8 @@ class Serial(base.Layer):
 
       weights.append(weights_or_cache_marker)
       states.append(state_or_cache_marker)
-    self.state = states
-    self.weights = weights
+    self.state = tuple(states)
+    self.weights = tuple(weights)
   # pylint: enable=protected-access
 
   def _n_inputs_n_outputs(self, layers):
@@ -151,7 +150,7 @@ class Parallel(base.Layer):
   Then Parallel(F, G, H) will take 6 inputs and give 4 outputs:
 
     - inputs: a, b, c, d, e, f
-    - outputs: F(a), G(b, c, d), h1, h2
+    - outputs: F(a), G(b, c, d), h1, h2     where h1, h2 = H(e, f)
 
   As an important special case, a None argument to Parallel acts as if it takes
   one argument, which it leaves unchanged. (It acts as a one-arg no-op.) For
@@ -385,20 +384,34 @@ class Scan(base.Layer):
     Scan(add)([1, 2, 3], 0) = [1, 3, 6], 6
   """
 
-  def __init__(self, layer, axis=0, n_carry=1, remat=False):
+  def __init__(self, layer, axis=0, n_carry=1, remat=False, mode='train'):
     super().__init__(n_in=layer.n_in, n_out=layer.n_out)
     self._sublayers = [layer]
     self._n_carry = n_carry
     self._axis = axis
     self._remat = remat
     self._weights = (None,)
-    self._state = (None,)
+    self._state = (None, ())
+    self._mode = mode
 
   @property
   def sublayer(self):
     """Returns the unique sublayer managed by this layer."""
     return self._sublayers[0]
 
+  @property
+  def state(self):
+    """Returns a tuple containing this layer's state."""
+    return (self.sublayer.state, self._state[1])
+
+  @state.setter
+  def state(self, state):
+    """Recursively sets state on this layer the sublayer."""
+    if isinstance(state, dict) and state == base.GET_STATE_FROM_CACHE:
+      return
+    self._state = (None, state[1])
+    self.sublayer.state = state[0]
+
   def forward(self, inputs):
     """Executes this layer as part of a forward pass through the model."""
     weights = self.weights[0]
@@ -418,13 +431,18 @@ class Scan(base.Layer):
 
     if n_carry > 0:
       xs = inputs[:-n_carry]  # Split input stack into inputs and carry.
-      init = (inputs[-n_carry:], self.state[0], jnp.array(0, dtype=jnp.int32))
+      xs_carry = inputs[-n_carry:]
+      if self._mode == 'predict' and self._state[1] is not ():  # pylint: disable=literal-comparison
+        xs_carry = self._state[1]
+      init = (xs_carry, self.state[0], jnp.array(0, dtype=jnp.int32))
     else:
+      xs_carry = ()
       xs, init = inputs, ([], self.state[0], jnp.array(0, dtype=jnp.int32))
     ys, (carry, new_state, _) = _scan(scannable_fn, xs, init,
                                       axis=self._axis, remat=self._remat)
     res = ys + carry if n_carry > 0 else ys
-    self.state = (new_state,)
+    state_carry = carry if self._mode == 'predict' and n_carry > 0 else ()
+    self.state = (new_state, state_carry)
     return res  # Put outputs and carry back on stack.
 
   def init_weights_and_state(self, input_signature):
@@ -439,7 +457,7 @@ class Scan(base.Layer):
         layer_sig = ShapeDtype(_shape_without_axis(input_signature, self._axis),
                                input_signature.dtype)
       weights, state = self.sublayer.init(layer_sig)
-      self.state = (state,)
+      self.state = (state, ())
       self.weights = (weights,)
     else:
       xs = input_signature[:-n_carry]
@@ -448,7 +466,7 @@ class Scan(base.Layer):
                    for x in xs]
       layer_signature = tuple(xs_slices + list(init))
       weights, state = self.sublayer.init(layer_signature, use_cache=True)
-      self.state = (state,)
+      self.state = (state, ())
       self.weights = (weights,)
 
 
@@ -909,9 +927,9 @@ class BatchLeadingAxes(base.Layer):
   """
 
   def __init__(self, layer, n_last_axes_to_keep=1):
-    if layer.n_in != 1 or layer.n_out != 1:
+    if layer.n_out != 1:
       raise ValueError('BatchLeadingAxes currently only works for layers with '
-                       f'n_in = n_out = 1, got {(layer.n_in, layer.n_out)}.')
+                       f'n_out = 1, got {layer.n_out}.')
     super().__init__(n_in=layer.n_in, n_out=layer.n_out)
     self._sublayers = [layer]
     self._n_last_axes_to_keep = n_last_axes_to_keep
@@ -925,22 +943,36 @@ class BatchLeadingAxes(base.Layer):
 
   def forward(self, inputs):
     """Executes this layer as part of a forward pass through the model."""
-    batched_axes_shape = list(inputs.shape[:-self._n_last_axes_to_keep])
-    batched_shape = [-1] + list(inputs.shape[-self._n_last_axes_to_keep:])
-    inputs = jnp.reshape(inputs, batched_shape)
+    if self._n_in == 1:
+      inputs = [inputs]
+    new_inputs = []
+    for old_input in inputs:
+      batched_axes_shape = list(old_input.shape[:-self._n_last_axes_to_keep])
+      batched_shape = [-1] + list(old_input.shape[-self._n_last_axes_to_keep:])
+      new_inputs.append(jnp.reshape(old_input, batched_shape))
+    new_inputs = tuple(new_inputs)
+    if self._n_in == 1:
+      new_inputs = new_inputs[0]
     res, layer_state = self.sublayer.pure_fn(
-        inputs, self.weights[0], self.state[0], self.rng)
+        new_inputs, self.weights[0], self.state[0], self.rng)
     self.state = (layer_state,)
     return jnp.reshape(res, batched_axes_shape + list(res.shape[1:]))
 
   def init_weights_and_state(self, input_signature):
     """Initializes weights and state for inputs with the given signature."""
-    batched_size = 1
-    for d in input_signature.shape[:-self._n_last_axes_to_keep]:
-      batched_size *= d
-    batched_shape = [batched_size] + list(
-        input_signature.shape[-self._n_last_axes_to_keep:])
-    batched_signature = ShapeDtype(batched_shape, input_signature.dtype)
+    if self._n_in == 1 and not isinstance(input_signature, (list, tuple)):
+      input_signature = (input_signature,)
+    batched_signature = []
+    for sub_input_signature in input_signature:
+      batched_size = 1
+      for d in sub_input_signature.shape[:-self._n_last_axes_to_keep]:
+        batched_size *= d
+      batched_shape = [batched_size] + list(
+          sub_input_signature.shape[-self._n_last_axes_to_keep:])
+      batched_signature.append(ShapeDtype(batched_shape,
+                                          sub_input_signature.dtype))
+    if self._n_in == 1:
+      batched_signature = batched_signature[0]
     weights, layer_state = self.sublayer.init(batched_signature, use_cache=True)
     self.state = (layer_state,)
     self.weights = (weights,)
