@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -60,7 +60,7 @@ class AxialPositionalEncoding(layer_base.Layer):
       assert self._dropout == 0.0
       emb = jnp.concatenate(embs, -1)
       emb = jnp.reshape(emb, (inputs.shape[0], -1, emb.shape[-1]))
-      emb = jax.lax.dynamic_slice_in_dim(emb, state, inputs.shape[1], axis=1)
+      emb = fastmath.dynamic_slice_in_dim(emb, state, inputs.shape[1], axis=1)
       self.state = state + inputs.shape[1]
       return inputs + emb
     elif self._dropout == 0:
@@ -104,11 +104,101 @@ class AxialPositionalEncoding(layer_base.Layer):
     self.weights = tuple(weights)
 
 
+class SinCosPositionalEncoding(layer_base.Layer):
+  """Implements the sin-cos positional encoding."""
+
+  def __init__(self, add_offset=2048, dropout=0.0, dropout_broadcast_dims=(-2,),
+               start_from_zero_one_in=2, mode='train'):
+    """Creates a SinCosPositionalEncoding instance.
+
+    Args:
+      add_offset: Maximumnumber to add to positions during training.
+      dropout: Probability of *not* adding positional encoding to a sequence
+          position.
+      dropout_broadcast_dims: Axes along which dropout mask values are
+          broadcast rather than individually set at random.
+      start_from_zero_one_in: how often to start from 0 during training,
+          every one in that many times (e.g., if 4, then it's 25% of the time).
+      mode: One of `'train'`, `'eval'`, or `'predict'`.
+    """
+    super().__init__()
+    self._add_offset = add_offset
+    if dropout >= 1.0:
+      raise ValueError('Dropout rates must be lower than 1.')
+    if mode == 'train':
+      self._dropout = dropout
+    else:
+      self._dropout = 0.0
+    self._dropout_broadcast_dims = dropout_broadcast_dims
+    self._start_from_zero_one_in = start_from_zero_one_in
+    self._mode = mode
+
+  def _sincos(self, start, length, d_feature):
+    """Create the sin-cos tensor of shape [1, length, d_feature]."""
+    position = jnp.arange(0, length)[:, None] + start
+    div_term = jnp.exp(
+        jnp.arange(0, d_feature, 2) * -(jnp.log(10000.0) / d_feature))
+    sin = jnp.sin(position * div_term)
+    cos = jnp.cos(position * div_term)
+    pe = jnp.concatenate([sin, cos], axis=1)
+    return pe[None, :, :]  # [1, length, d_feature]
+
+  def forward(self, inputs):
+    """Returns the input activations, with added positional information."""
+    if self._mode != 'predict':
+      x = inputs
+      length = jnp.shape(x)[1]
+      if self._mode != 'train':
+        start = 0
+      else:
+        rng1, rng2 = fastmath.random.split(self.rng, 2)
+        start = fastmath.random.randint(rng1, (), 0, self._add_offset)
+        start_from_nonzero = fastmath.random.randint(
+            rng2, (), 0, self._start_from_zero_one_in)
+        start_from_nonzero = jnp.minimum(1, start_from_nonzero)
+        start *= start_from_nonzero
+      px = self._sincos(start, length, inputs.shape[2])
+      if self._dropout == 0:
+        return x + px
+      else:
+        noise_shape = list(px.shape)
+        for dim in self._dropout_broadcast_dims:
+          noise_shape[dim] = 1
+        keep_prob = 1.0 - self._dropout
+        keep = fastmath.random.bernoulli(self.rng, keep_prob,
+                                         tuple(noise_shape))
+        multiplier = keep.astype(x.dtype) / keep_prob
+        return x + px * multiplier
+    else:
+      if self._dropout != 0:
+        raise ValueError(f'In predict mode, but dropout rate '
+                         f'({self._dropout}) is not zero.')
+
+      # State in this class is only used for fast inference. In that case,
+      # the model is called with consecutive elements position-by-position.
+      # This positional encoding layer needs to store the index of the current
+      # position then and increment it on each call -- that's how state is used
+      # and updated below.
+      pe = self._sincos(self.state, inputs.shape[1], inputs.shape[2])
+      self.state += inputs.shape[1]
+      return inputs + pe
+
+  def init_weights_and_state(self, input_signature):
+    """Randomly initializes the positional encoding vectors.
+
+    Args:
+      input_signature: `ShapeDtype` instance characterizing the input this
+          layer should compute on.
+    """
+    if self._mode == 'predict':
+      self.state = jnp.zeros((), dtype=jnp.int32)
+
+
 class FixedBasePositionalEncoding(layer_base.Layer):
   """Implements fixed-base positional encoding."""
 
   def __init__(self, bases=[11, 13, 14, 15], n_digits=8,  #  pylint: disable=dangerous-default-value
-               start_from_zero_one_in=100, base_dropout_one_in=100,
+               start_from_zero_one_in=2, base_dropout_one_in=100,
                mode='train', initializer=init.RandomUniformInitializer(1e-4)):
     super().__init__()
     self._bases = bases
@@ -120,11 +210,15 @@ class FixedBasePositionalEncoding(layer_base.Layer):
 
   def forward(self, x):
     rng = self.rng
+    base_weights, start_vec = self.weights
     batch_size, length = x.shape[0], x.shape[1]
     max_pos = min(self._bases)**self._n_digits
     rng1, rng2, rng3 = fastmath.random.split(rng, 3)
     assert length < max_pos, 'length (%d) >= max_pos (%d)' % (length, max_pos)
     positions = jnp.arange(0, length)[None, :]
+    # In training we'll randomize starts for better generalization.
+    # We use the trainable start_vec to compensate and give model a way
+    # to learn what is the starting position in a sequence.
     if self._mode == 'train':
       # In 1% of training cases still start from 0 to be exactly as in eval.
       start_from_nonzero = fastmath.random.randint(
@@ -134,6 +228,8 @@ class FixedBasePositionalEncoding(layer_base.Layer):
           rng2, (batch_size,), 0, max_pos-length)
       random_start *= start_from_nonzero
       positions += random_start[:, None]
+    if self._mode == 'predict':
+      positions += self.state
     res = []
     for bn, base in enumerate(self._bases):
       pos_embeddings = []
@@ -141,7 +237,7 @@ class FixedBasePositionalEncoding(layer_base.Layer):
       for i in range(self._n_digits):
         cur_indices = jnp.mod(cur_positions, base)
         cur_positions = cur_positions // base
-        s = self.weights[bn][i]
+        s = base_weights[bn][i]
         pos_embeddings.append(cur_indices.astype(jnp.float32)[:, :, None] * s)
       embeddings = jnp.concatenate(pos_embeddings, axis=-1)
       if self._mode == 'train':
@@ -150,16 +246,31 @@ class FixedBasePositionalEncoding(layer_base.Layer):
         base_dropout = jnp.minimum(1, base_dropout).astype(jnp.float32)
         embeddings *= base_dropout[:, None, None]
       res.append(embeddings)
-    res = sum(res) + jnp.zeros_like(x)
+    res = sum(res)  # Sum embeddings from all bases.
+    # Add start_vec to the first position only to mark it as starting.
+    res0 = res[:, 0, :][:, None, :]
+    start_pos = res0 + start_vec
+    if self._mode == 'predict':
+      start_pos = jnp.where(jnp.equal(self.state, 0), start_pos, res0)
+      self.state += length  # Add input length to state.
+    res = jnp.concatenate([start_pos, res[:, 1:, :]], axis=1)
     return x + res
 
   def init_weights_and_state(self, input_signature):
     d_feature = input_signature.shape[-1]
-    assert d_feature % self._n_digits == 0
+    if d_feature % self._n_digits != 0:
+      raise ValueError(
+          f'd_feature({d_feature}) % self._n_digits({self._n_digits}) != 0')
     d_weight = d_feature // self._n_digits
-    self.weights = [[self._initializer((1, d_weight), rng)
-                     for rng in fastmath.random.split(self.rng, self._n_digits)]
+    rng1, rng2 = fastmath.random.split(self.rng, 2)
+    base_weights = [[self._initializer((1, d_weight), rng)
+                     for rng in fastmath.random.split(rng1, self._n_digits)]
                     for _ in self._bases]
+    # Special vector to mark the starting position.
+    start_vec = self._initializer((1, 1, d_feature), rng2)
+    self.weights = (base_weights, start_vec)
+    if self._mode == 'predict':
+      self.state = jnp.zeros((), dtype=jnp.int32)
 
 
 def threefry_2x32_prf(key, x: jnp.ndarray) -> jnp.ndarray:
