@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -26,20 +26,13 @@ from trax.models import mlp
 
 class MLPTest(absltest.TestCase):
 
-  def test_pure_mlp_forward_shape(self):
-    model = mlp.PureMLP(layer_widths=(32, 16, 8))
+  def test_mlp_forward_shape(self):
+    model = mlp.MLP(layer_widths=(32, 16, 8))
     x = np.ones((7, 28, 28, 3)).astype(np.float32)
     _, _ = model.init(shapes.signature(x))
     y = model(x)
     self.assertEqual(y.shape, (7, 8))
 
-  def test_mlp_forward_shape(self):
-    model = mlp.MLP(d_hidden=32, n_output_classes=10)
-    x = np.ones((3, 28, 28, 1)).astype(np.float32)
-    _, _ = model.init(shapes.signature(x))
-    y = model(x)
-    self.assertEqual(y.shape, (3, 10))
-
 
 
 if __name__ == '__main__':
