@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -18,8 +18,8 @@
 
 from trax import layers as tl
 from trax.fastmath import numpy as jnp
-from trax.models.research import configurable_transformer
-from trax.models.research import transformer_no_enc_dec_attention as t2
+from trax.models.research import configurable_transformer as ct
+from trax.models.research import transformer2 as t2
 
 
 # Layers are always CamelCase, but functions in general are snake_case
@@ -29,7 +29,9 @@ from trax.models.research import transformer_no_enc_dec_attention as t2
 def DecoderBlock(d_model, d_ff, d_attention_key, d_attention_value,
                  n_heads, attention_type, dropout, ff_activation,
                  ff_dropout, ff_use_sru, ff_chunk_size, ff_sparsity,
-                 attention_chunk_size, mode):
+                 attention_chunk_size, n_attention_layers=1,
+                 n_feedforward_layers=1, center_layernorm=True,
+                 use_bfloat16=False, mode='train'):
   """Reversible transformer decoder layer.
 
   Args:
@@ -46,55 +48,41 @@ def DecoderBlock(d_model, d_ff, d_attention_key, d_attention_value,
     ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
     ff_sparsity: int, if > 0 use sparse feed-forward block with this sparsity
     attention_chunk_size: int, if > 0 run attention chunked at this size
+    n_attention_layers: how many residual causal attention layers should we
+      have before the feed-forward block (default: 1, the standard block)
+    n_feedforward_layers: how many FFNN layers should we have (default 1).
+    center_layernorm: whether to use centering in LayerNorm (default) or if
+      to skip it, which is known as RMS normalization.
+    use_bfloat16: whether to use bfloat16 for weights (default: False).
     mode: str: 'train' or 'eval'
 
 
   Returns:
     the layer.
   """
-  attention = configurable_transformer.ApplyAttentionLayer(
-      attention_type, d_model, n_heads, d_attention_key, d_attention_value,
-      True, False, dropout, dropout, attention_chunk_size, mode)
-  attention_half_residual = tl.ReversibleHalfResidual(
-      tl.LayerNorm(),
-      attention_layer=attention,
-  )
+  # pylint: disable=g-complex-comprehension
+  attention_half_residuals = [
+      [tl.ReversibleHalfResidual(
+          tl.LayerNorm(center=center_layernorm),
+          attention_layer=ct.ApplyAttentionLayer(
+              attention_type, d_model, n_heads, d_attention_key,
+              d_attention_value, True, False, dropout, dropout,
+              attention_chunk_size, mode),
+          name='ReversibleHalfResidualDecoderAttn'),
+       tl.ReversibleSwap()
+      ] for _ in range(n_attention_layers)]
 
-  feed_forward = configurable_transformer.FeedForwardWithOptions(
-      d_model, d_ff, dropout, [-2], ff_activation, ff_dropout,
-      ff_chunk_size, ff_use_sru, ff_sparsity, mode)
-
-  return [
-      attention_half_residual,
-      tl.ReversibleSwap(),
-      tl.ReversibleHalfResidual(feed_forward),
-      tl.ReversibleSwap(),
-  ]
-
-
-def PositionalEncoding(mode, dropout=None, max_len=None,
-                       axial_pos_shape=None, d_axial_pos_embs=None):
-  """Returns the positional encoding layer depending on the arguments."""
-  if not axial_pos_shape:
-    positional_encoding = tl.PositionalEncoding(
-        max_len=max_len, dropout=dropout, mode=mode)
-  elif axial_pos_shape == 'fixed-base':  # TODO(lukaszkaiser): remove this HACK
-    positional_encoding = tl.FixedBasePositionalEncoding(mode=mode)
-  elif axial_pos_shape == 'infinite':  # TODO(lukaszkaiser): remove this HACK
-    positional_encoding = tl.InfinitePositionalEncoding(affine=False)
-  elif axial_pos_shape == 'infinite-affine':
-    # TODO(lukaszkaiser): remove this HACK
-    positional_encoding = tl.InfinitePositionalEncoding()
-  elif axial_pos_shape == 'time-bin':  # TODO(lukaszkaiser): remove this HACK
-    positional_encoding = tl.TimeBinPositionalEncoding()
-  else:
-    assert d_axial_pos_embs is not None
-    positional_encoding = tl.AxialPositionalEncoding(
-        shape=axial_pos_shape, d_embs=d_axial_pos_embs,
-        dropout_broadcast_dims=tuple(range(1, len(axial_pos_shape) + 1)),
-        dropout=dropout, mode=mode)
-
-  return positional_encoding
+  feed_forwards = [
+      [tl.ReversibleHalfResidual(
+          ct.FeedForwardWithOptions(
+              d_model, d_ff, dropout, [-2], ff_activation, ff_dropout,
+              ff_chunk_size, ff_use_sru, ff_sparsity, center_layernorm,
+              mode, use_bfloat16),
+          name='ReversibleHalfResidualDecoderFF'),
+       tl.ReversibleSwap()
+      ] for _ in range(n_feedforward_layers)]
+  # pylint: enable=g-complex-comprehension
+  return attention_half_residuals + feed_forwards
 
 
 def ReformerLM(vocab_size,
@@ -107,12 +95,17 @@ def ReformerLM(vocab_size,
                dropout=0.1,
                max_len=2048,
                attention_type=tl.SelfAttention,
-               axial_pos_shape=(),
-               d_axial_pos_embs=None,
+               pos_type=None,
+               pos_axial_shape=(),
+               pos_d_axial_embs=None,
                ff_activation=tl.FastGelu,
                ff_use_sru=0,
                ff_chunk_size=0,
                ff_sparsity=0,
+               loss_sparsity_type='mult',
+               loss_sparsity=0,
+               loss_d_lowrank=0,
+               loss_sparsity_prob=None,
                attention_chunk_size=0,
                mode='train'):
   """Reversible transformer language model (only uses a decoder, no encoder).
@@ -128,22 +121,29 @@ def ReformerLM(vocab_size,
     dropout: float: dropout rate (how much to drop out)
     max_len: int: maximum symbol length for positional encoding
     attention_type: class: attention class to use, such as SelfAttention.
-    axial_pos_shape: tuple of ints: input shape to use for the axial position
+    pos_type: string, the type of positional embeddings to use.
+    pos_axial_shape: tuple of ints: input shape to use for the axial position
       encoding. If unset, axial position encoding is disabled.
-    d_axial_pos_embs: tuple of ints: depth of position embedding for each axis.
-      Tuple length must match axial_pos_shape, and values must sum to d_model.
+    pos_d_axial_embs: tuple of ints: depth of position embedding for each axis.
+      Tuple length must match pos_axial_shape, and values must sum to d_model.
     ff_activation: the non-linearity in feed-forward layer
     ff_use_sru: int; if > 0, we use this many SRU layers instead of feed-forward
     ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
     ff_sparsity: int, if > 0 use sparse feed-forward block with this sparsity
+    loss_sparsity_type: str, type of sparsity to used in loss layer. See
+      SparseDenseWithOptions for options. None if no sparsity should be used.
+    loss_sparsity: int, the sparsity for loss layer (if used)
+    loss_d_lowrank: int, the dimensions for intermediate layer (if used)
+    loss_sparsity_prob: float, the probability for sparse version of loss to be
+      used. If None, only sparse version is used.
     attention_chunk_size: int, if > 0 run attention chunked at this size
     mode: str: 'train', 'eval', or 'predict'
 
   Returns:
     the layer.
   """
-  positional_encoding = PositionalEncoding(
-      mode, dropout, max_len, axial_pos_shape, d_axial_pos_embs)
+  positional_encoding = ct.PositionalEncoder(
+      mode, dropout, max_len, pos_type, pos_axial_shape, pos_d_axial_embs)
 
   positional_embedder = [
       tl.Embedding(vocab_size, d_model),
@@ -172,6 +172,15 @@ def ReformerLM(vocab_size,
         mode=mode)
     decoder_blocks.append(decoder_block)
 
+  dense_loss_layer = tl.SparseDenseWithOptions(
+      vocab_size,
+      d_input=d_model,
+      sparsity_type=loss_sparsity_type,
+      sparsity=loss_sparsity,
+      d_lowrank=loss_d_lowrank,
+      prob_sparse=loss_sparsity_prob,
+      mode=mode)
+
   return tl.Serial(
       tl.ShiftRight(mode=mode),
       positional_embedder,
@@ -182,8 +191,7 @@ def ReformerLM(vocab_size,
       # LayerNorm, and whether dropout broadcasting is needed here.
       tl.LayerNorm(),
       tl.Dropout(rate=dropout, shared_axes=[-2], mode=mode),  # pylint: disable=no-value-for-parameter
-      tl.Dense(vocab_size),
-      tl.LogSoftmax(),
+      dense_loss_layer,
   )
 
 
@@ -199,8 +207,9 @@ def ReformerShortenLM(vocab_size,
                       dropout=0.1,
                       max_len=2048,
                       attention_type=tl.SelfAttention,
-                      axial_pos_shape=(),
-                      d_axial_pos_embs=None,
+                      pos_type=None,
+                      pos_axial_shape=(),
+                      pos_d_axial_embs=None,
                       ff_activation=tl.FastGelu,
                       ff_use_sru=0,
                       ff_chunk_size=0,
@@ -232,10 +241,11 @@ def ReformerShortenLM(vocab_size,
     dropout: float: dropout rate (how much to drop out)
     max_len: int: maximum symbol length for positional encoding
     attention_type: class: attention class to use, such as SelfAttention.
-    axial_pos_shape: tuple of ints: input shape to use for the axial position
+    pos_type: string, the type of positional embeddings to use.
+    pos_axial_shape: tuple of ints: input shape to use for the axial position
       encoding. If unset, axial position encoding is disabled.
-    d_axial_pos_embs: tuple of ints: depth of position embedding for each axis.
-      Tuple length must match axial_pos_shape, values must sum to d_embedding.
+    pos_d_axial_embs: tuple of ints: depth of position embedding for each axis.
+      Tuple length must match pos_axial_shape, values must sum to d_embedding.
     ff_activation: the non-linearity in feed-forward layer
     ff_use_sru: int; if > 0, we use this many SRU layers instead of feed-forward
     ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
@@ -248,15 +258,8 @@ def ReformerShortenLM(vocab_size,
   """
   assert mode != 'predict'  # TODO(lukaszkaiser,kitaev): fast inference
 
-  if not axial_pos_shape:
-    positional_encoding = tl.PositionalEncoding(
-        max_len=max_len, dropout=dropout, mode=mode)
-  else:
-    assert d_axial_pos_embs is not None
-    positional_encoding = tl.AxialPositionalEncoding(
-        shape=axial_pos_shape, d_embs=d_axial_pos_embs,
-        dropout_broadcast_dims=tuple(range(1, len(axial_pos_shape) + 1)),
-        dropout=dropout, mode=mode)
+  positional_encoding = ct.PositionalEncoder(
+      mode, dropout, max_len, pos_type, pos_axial_shape, pos_d_axial_embs)
 
   positional_embedder = [
       tl.Embedding(vocab_size, d_embedding),
@@ -315,14 +318,15 @@ def ReformerShortenLM(vocab_size,
       tl.Relu(),
       tl.SRU(d_embedding),  # One RNN layer for conditional dependence.
       tl.Dense(vocab_size),
-      tl.LogSoftmax()
   )
   # pylint: enable=g-long-lambda
 
 
 def EncoderBlock(d_model, d_ff, n_heads, attention_type, dropout, ff_activation,
                  ff_dropout, ff_use_sru=0, ff_chunk_size=0, ff_sparsity=0,
-                 attention_chunk_size=0, mode='train'):
+                 attention_chunk_size=0, center_layernorm=True,
+                 use_bfloat16=False, use_two_swaps_per_block=True,
+                 mode='train'):
   """Returns a list of layers that implements a Reformer encoder block.
 
   The input to the layer is a pair, (activations, mask), where the mask was
@@ -341,6 +345,11 @@ def EncoderBlock(d_model, d_ff, n_heads, attention_type, dropout, ff_activation,
     ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
     ff_sparsity: int, if > 0 use sparse feed-forward block with this sparsity
     attention_chunk_size: int, if > 0 run attention chunked at this size
+    center_layernorm: whether to use centering in LayerNorm (default) or if
+      to skip it, which is known as RMS normalization.
+    use_bfloat16: whether to use bfloat16 for weights (default: False)
+    use_two_swaps_per_block: bool, if True use two reversible swaps in Encoder
+      block, otherwise use only one swap.
     mode: str: 'train' or 'eval'
 
   Returns:
@@ -352,30 +361,48 @@ def EncoderBlock(d_model, d_ff, n_heads, attention_type, dropout, ff_activation,
     # to 'eval' mode instead.
     mode = 'eval'
 
-  attention = configurable_transformer.ApplyAttentionLayer(
+  attention = ct.ApplyAttentionLayer(
       attention_type=attention_type, d_model=d_model, n_heads=n_heads,
       d_qk=d_model//n_heads, d_v=d_model//n_heads, masked=True, causal=False,
       attention_dropout=dropout, output_dropout=dropout,
       attention_chunk_size=attention_chunk_size, mode=mode)
+  # TODO(lukaszkaiser): refactor efficient attention layers to unify the API
+  # If we're using standard attention, we need to pass reshaped mask and not
+  # return the mask to be compatible with the EfficientAttention API.
+  if attention.n_out == 2:
+    def reshape_mask(mask):
+      return jnp.reshape(mask, (mask.shape[0], 1, 1, mask.shape[1]))
+    attention = tl.Serial(
+        tl.Fn('ReshapeMask', lambda x, y: (x, reshape_mask(y)), n_out=2),
+        attention,
+        tl.Select([0], n_in=2)
+    )
+
   attention_half_residual = tl.ReversibleHalfResidual(
-      tl.LayerNorm(),
+      tl.LayerNorm(center=center_layernorm),
       attention_layer=attention,
+      name='ReversibleHalfResidualEncoderAttn'
   )
 
-  feed_forward = configurable_transformer.FeedForwardWithOptions(
+  feed_forward = ct.FeedForwardWithOptions(
       d_model, d_ff, dropout, [-2], ff_activation, ff_dropout,
-      ff_chunk_size, ff_use_sru, ff_sparsity, mode)
+      ff_chunk_size, ff_use_sru, ff_sparsity, center_layernorm,
+      mode, use_bfloat16)
 
-  return [
+  encoder_block = [
       attention_half_residual,
       tl.ReversibleSwap(),
-      tl.ReversibleHalfResidual(feed_forward),
-      tl.ReversibleSwap(),
+      tl.ReversibleHalfResidual(feed_forward,
+                                name='ReversibleHalfResidualEncoderFF'),
   ]
+  if use_two_swaps_per_block:
+    encoder_block.append(tl.ReversibleSwap())
+  return encoder_block
 
 
 def EncoderDecoderBlock(d_model, d_ff, n_heads, dropout, ff_activation,
-                        ff_dropout, mode):
+                        ff_dropout, mode, ff_use_sru=0, ff_chunk_size=0,
+                        ff_sparsity=0):
   """Reversible transformer decoder layer.
 
   Args:
@@ -386,6 +413,9 @@ def EncoderDecoderBlock(d_model, d_ff, n_heads, dropout, ff_activation,
     ff_activation: the non-linearity in feed-forward layer
     ff_dropout: float: (optional) separate dropout rate for feed-forward layer
     mode: str: 'train' or 'eval'
+    ff_use_sru: int; if > 0, we use this many SRU layers instead of feed-forward
+    ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
+    ff_sparsity: int, if > 0 use sparse feed-forward block with this sparsity
 
   Returns:
     the layer.
@@ -409,8 +439,9 @@ def EncoderDecoderBlock(d_model, d_ff, n_heads, dropout, ff_activation,
       attention_layer=causal_attention,
   )
 
-  feed_forward = configurable_transformer.FeedForward(
-      d_model, d_ff, dropout, ff_activation, ff_dropout, mode)
+  feed_forward = ct.FeedForwardWithOptions(
+      d_model, d_ff, dropout, [-2], ff_activation, ff_dropout,
+      ff_chunk_size, ff_use_sru, ff_sparsity, True, mode)
 
   return [                             # vec_d1 vec_d2 vec_e masks
       causal_attention_half_residual,
@@ -433,7 +464,13 @@ def Reformer(input_vocab_size,
              max_len=2048,
              ff_activation=tl.Relu,
              ff_dropout=None,
-             mode='train'):
+             mode='train',
+             pos_type=None,
+             pos_axial_shape=None,
+             pos_d_axial_embs=None,
+             ff_use_sru=0,
+             ff_chunk_size=0,
+             ff_sparsity=0):
   """Reversible transformer encoder-decoder model.
 
   This model expects an input pair: target, source.
@@ -456,35 +493,39 @@ def Reformer(input_vocab_size,
     ff_dropout: float: (optional) separate dropout rate at feed-forward
       nonlinearity. This is called relu_dropout in T2T.
     mode: str: 'train' or 'eval'
+    pos_type: string, the type of positional embeddings to use.
+    pos_axial_shape: tuple of ints: input shape to use for the axial position
+      encoding. If unset, axial position encoding is disabled.
+    pos_d_axial_embs: tuple of ints: depth of position embedding for each axis.
+      Tuple length must match pos_axial_shape, and values must sum to d_model.
+    ff_use_sru: int; if > 0, we use this many SRU layers instead of feed-forward
+    ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
+    ff_sparsity: int, if > 0 use sparse feed-forward block with this sparsity
 
   Returns:
     A Reformer model as a layer that maps from a target, source pair to
     activations over a vocab set.
   """
-  def PositionalEncoder(vocab_size, mode):  # tokens --> vectors
-    # TODO(kitaev): axial positional encoding is better for very long sequences.
-    positional_encoding = tl.PositionalEncoding(
-        max_len=max_len, dropout=dropout, mode=mode)
-    return [
-        tl.Embedding(vocab_size, d_model),
-        tl.Dropout(rate=dropout, shared_axes=[-2], mode=mode),
-        positional_encoding,
-    ]
-
-  # Mode 'predict' means that the decoder should be run one token at a time.
-  # The encoder only ever runs over full sequences, which is why it's switched
-  # to 'eval' mode instead.
-  in_encoder = PositionalEncoder(
-      input_vocab_size, mode='eval' if mode == 'predict' else mode)
-  if output_vocab_size is None:
-    output_vocab_size = input_vocab_size
-  out_encoder = PositionalEncoder(output_vocab_size, mode)
+  in_encoder, out_encoder, output_vocab_size = (
+      ct.EmbeddingAndPositionalEncodings(
+          input_vocab_size,
+          d_model,
+          mode,
+          dropout,
+          [-2],  # dropout_shared_axes
+          max_len,
+          output_vocab_size=output_vocab_size,
+          pos_type=pos_type,
+          pos_axial_shape=pos_axial_shape,
+          pos_d_axial_embs=pos_d_axial_embs)
+  )
 
   # pylint: disable=g-complex-comprehension
   encoder_blocks = [
       EncoderBlock(
           d_model, d_ff, n_heads, tl.SelfAttention, dropout, ff_activation,
-          ff_dropout, mode=mode)
+          ff_dropout, mode=mode, ff_use_sru=ff_use_sru,
+          ff_chunk_size=ff_chunk_size, ff_sparsity=ff_sparsity)
       for _ in range(n_encoder_layers)]
   # pylint: enable=g-complex-comprehension
 
@@ -498,10 +539,14 @@ def Reformer(input_vocab_size,
   if mode == 'predict':
     encoder = tl.Cache(encoder)
 
+  # pylint: disable=g-complex-comprehension
   encoder_decoder_blocks = [
       EncoderDecoderBlock(
-          d_model, d_ff, n_heads, dropout, ff_activation, ff_dropout, mode)
+          d_model, d_ff, n_heads, dropout, ff_activation, ff_dropout, mode,
+          ff_use_sru=ff_use_sru, ff_chunk_size=ff_chunk_size,
+          ff_sparsity=ff_sparsity)
       for _ in range(n_decoder_layers)]
+  # pylint: enable=g-complex-comprehension
 
   # Assemble and return the model.
   return tl.Serial(
@@ -529,7 +574,6 @@ def Reformer(input_vocab_size,
       # Map to output vocab.
       tl.Select([0], n_in=3),               # vec_d .....
       tl.Dense(output_vocab_size),          # vec_d .....
-      tl.LogSoftmax(),                      # vec_d .....
   )
 
 
@@ -546,15 +590,30 @@ def Reformer2(input_vocab_size,
               max_len=2048,
               encoder_attention_type=tl.SelfAttention,
               encoder_decoder_attention_type=tl.SelfAttention,
-              axial_pos_shape='fixed-base',
-              d_axial_pos_embs=None,
+              pos_type='fixed-base',
+              pos_axial_shape=(),
+              pos_d_axial_embs=None,
+              pos_start_from_zero_prob=1.0,
+              pos_max_offset_to_add=0,
               ff_activation=tl.Relu,
               ff_use_sru=0,
               ff_chunk_size=0,
               ff_dropout=None,
               ff_sparsity=0,
+              loss_sparsity_type='mult',
+              loss_sparsity=0,
+              loss_d_lowrank=0,
+              loss_sparsity_prob=None,
               attention_chunk_size=0,
               n_layers_forget=0,
+              forget_dense=True,
+              n_decoder_attention_layers=2,
+              use_bfloat16=False,
+              reversible_encoder=False,
+              use_two_swaps_per_encoder_block=True,
+              center_layernorm=True,
+              half_before_layer=None,
+              double_after_layer=None,
               mode='train'):
   """Reversible transformer encoder-decoder model.
 
@@ -579,18 +638,40 @@ def Reformer2(input_vocab_size,
     encoder_attention_type: class: attention class to use, such as SelfAttention
     encoder_decoder_attention_type: class: attention class to use, such as
       SelfAttention
-    axial_pos_shape: tuple of ints: input shape to use for the axial position
+    pos_type: string, the type of positional embeddings to use.
+    pos_axial_shape: tuple of ints: input shape to use for the axial position
       encoding. If unset, axial position encoding is disabled.
-    d_axial_pos_embs: tuple of ints: depth of position embedding for each axis.
-      Tuple length must match axial_pos_shape, and values must sum to d_model.
+    pos_d_axial_embs: tuple of ints: depth of position embedding for each axis.
+      Tuple length must match pos_axial_shape, and values must sum to d_model.
+    pos_start_from_zero_prob: how often to start from 0 during training,
+          (if 1.0, we always start from position 0, if less, we randomize).
+    pos_max_offset_to_add: maximum offset to add to positions during training
+        when randomizing; this offset plus input length must still be less than
+        max_len for all training examples.
     ff_activation: the non-linearity in feed-forward layer
     ff_use_sru: int; if > 0, we use this many SRU layers instead of feed-forward
     ff_chunk_size: int; if > 0, chunk feed-forward into this-sized chunks
     ff_dropout: float: (optional) separate dropout rate at feed-forward
       nonlinearity. This is called relu_dropout in T2T.
     ff_sparsity: int, if > 0 use sparse feed-forward block with this sparsity
+    loss_sparsity_type: str, type of sparsity to used in loss layer. See
+      SparseDenseWithOptions for options. None if no sparsity should be used.
+    loss_sparsity: int, the sparsity for loss layer (if used)
+    loss_d_lowrank: int, the dimensions for intermediate layer (if used)
+    loss_sparsity_prob: float, the probability for sparse version of loss to be
+      used. If None, only sparse version is used.
     attention_chunk_size: int, if > 0 run attention chunked at this size
     n_layers_forget: how often to have a forgetting block between layers
+    forget_dense: whether to use Dense or no-op (Serial) as a forget layer.
+    n_decoder_attention_layers: how many attention layers in a decoder block
+    use_bfloat16: whether to use bfloat16 for weights (default: False)
+    reversible_encoder: whether to be reversible through the encoder
+    use_two_swaps_per_encoder_block: whether to allow even number of swaps in
+      the encoder
+    center_layernorm: whether to use centering in LayerNorm (default) or if
+      to skip it, which is known as RMS normalization.
+    half_before_layer: int, half d_model and d_ff before that layer
+    double_after_layer: int, double d_model and d_ff after that layer
     mode: str: 'train' or 'eval'
 
   Returns:
@@ -598,43 +679,51 @@ def Reformer2(input_vocab_size,
     activations over a vocab set.
   """
   # Set default dimensions for attention head key and value sizes.
+  if (d_model / 2) % n_heads != 0:
+    raise ValueError(f'n_heads ({n_heads}) must divide d_model/2 ({d_model/2})')
   if d_attention_key is None:
-    if d_model % n_heads != 0:
-      raise ValueError(f'n_heads ({n_heads}) must divide d_model ({d_model})')
     d_attention_key = d_model // n_heads
   if d_attention_value is None:
-    if d_model % n_heads != 0:
-      raise ValueError(f'n_heads ({n_heads}) must divide d_model ({d_model})')
     d_attention_value = d_model // n_heads
 
+  # Set values of d_model, d_ff and d_qkv for the first stage.
+  d_model1, d_ff1 = d_model, d_ff
+  d_attention_key1, d_attention_value1 = d_attention_key, d_attention_value
+  if half_before_layer:
+    d_model1, d_ff1 = d_model / 2, d_ff / 2
+    d_attention_key1 = d_attention_key / 2
+    d_attention_value1 = d_attention_value / 2
+
+  # Set values of d_model, d_ff and d_qkv for the final stage.
+  d_model2, d_ff2 = d_model, d_ff
+  d_attention_key2, d_attention_value2 = d_attention_key, d_attention_value
+  if double_after_layer:
+    d_model2, d_ff2 = d_model * 2, d_ff * 2
+    d_attention_key2 = d_attention_key * 2
+    d_attention_value2 = d_attention_value * 2
+
   # Vector embeddings.
-  def Embedder(vocab_size):  # tokens --> vectors
-    return [
-        tl.Embedding(vocab_size, d_model),
-        tl.Dropout(rate=dropout, shared_axes=[-2], mode=mode),
-    ]
-
-  in_embedder = Embedder(input_vocab_size)
-  out_embedder = (in_embedder if output_vocab_size is None
-                  else Embedder(output_vocab_size))
-
-  def PositionalEnc(mode):
-    return PositionalEncoding(
-        mode, dropout, max_len, axial_pos_shape, d_axial_pos_embs)
-
-  # Mode 'predict' means that the decoder should be run one token at a time.
-  # The encoder only ever runs over full sequences, which is why it's switched
-  # to 'eval' mode instead.
-  encoder_mode = 'eval' if mode == 'predict' else mode
-  in_encoder = in_embedder + [PositionalEnc(encoder_mode)]
-  out_encoder = out_embedder + [PositionalEnc(mode)]
-  if output_vocab_size is None:
-    output_vocab_size = input_vocab_size
+  in_encoder, out_encoder, output_vocab_size = (
+      ct.EmbeddingAndPositionalEncodings(
+          input_vocab_size,
+          d_model1,
+          mode,
+          dropout,
+          [-2],  # dropout_shared_axes
+          max_len,
+          output_vocab_size=output_vocab_size,
+          pos_type=pos_type,
+          pos_axial_shape=pos_axial_shape,
+          pos_d_axial_embs=pos_d_axial_embs,
+          pos_start_from_zero_prob=pos_start_from_zero_prob,
+          pos_max_offset_to_add=pos_max_offset_to_add,
+          use_bfloat16=use_bfloat16)
+  )
 
   # pylint: disable=g-complex-comprehension
   encoder_blocks = [
       EncoderBlock(
-          d_model, d_ff, n_heads, encoder_attention_type,
+          d_model1, d_ff1, n_heads, encoder_attention_type,
           dropout=dropout,
           ff_activation=ff_activation,
           ff_dropout=ff_dropout,
@@ -642,17 +731,25 @@ def Reformer2(input_vocab_size,
           ff_chunk_size=ff_chunk_size,
           ff_sparsity=ff_sparsity,
           attention_chunk_size=attention_chunk_size,
+          center_layernorm=center_layernorm,
+          use_bfloat16=use_bfloat16,
+          use_two_swaps_per_block=use_two_swaps_per_encoder_block,
           mode=mode)
       for _ in range(n_encoder_layers)]
   # pylint: enable=g-complex-comprehension
 
-  encoder = tl.Serial([                # vec_e mask_e tok_e tok_d tok_d
-      tl.Dup(),                        # vec_e1 vec_e2 mask_e tok_e tok_d tok_d
-      _ReversibleSerialForget(encoder_blocks, d_model, n_layers_forget),
-      tl.Fn('XYAvg', lambda x, y: (x + y) / 2.0),
-      tl.Dense(d_model),
-      tl.LayerNorm(),
-  ])
+  encoder = [     # vec_e mask_e tok_e tok_d tok_d
+      tl.ReversibleSelect([0, 0]),     # vec_e1 vec_e2 mask_e tok_e tok_d tok_d
+      _ReversibleSerialForget(encoder_blocks, d_model1, n_layers_forget,
+                              forget_dense)
+  ]
+  if not reversible_encoder:
+    encoder += [
+        tl.Fn('XYAvg', lambda x, y: (x + y) / 2.0),
+        tl.Dense(d_model1, use_bfloat16=use_bfloat16),
+        tl.LayerNorm(),
+    ]
+  encoder = tl.Serial(encoder)
   if mode == 'predict':
     encoder = tl.Cache(encoder)
 
@@ -665,8 +762,14 @@ def Reformer2(input_vocab_size,
   for layer_idx in range(n_decoder_layers):
     layer_attention_type = encoder_decoder_attention_type[
         layer_idx % len(encoder_decoder_attention_type)]
+    # Grow d_model, d_ff, and d_qkv if requested.
+    d_m, d_f, d_k, d_v = d_model1, d_ff1, d_attention_key1, d_attention_value1
+    if half_before_layer and layer_idx >= half_before_layer:
+      d_m, d_f, d_k, d_v = d_model, d_ff, d_attention_key, d_attention_value
+    if double_after_layer and layer_idx > double_after_layer:
+      d_m, d_f, d_k, d_v = d_model2, d_ff2, d_attention_key2, d_attention_value2
     decoder_block = DecoderBlock(
-        d_model, d_ff, d_attention_key, d_attention_value, n_heads,
+        d_m, d_f, d_k, d_v, n_heads,
         attention_type=layer_attention_type,
         dropout=dropout,
         ff_activation=ff_activation,
@@ -675,8 +778,38 @@ def Reformer2(input_vocab_size,
         ff_chunk_size=ff_chunk_size,
         ff_sparsity=ff_sparsity,
         attention_chunk_size=attention_chunk_size,
+        n_attention_layers=n_decoder_attention_layers,
+        center_layernorm=center_layernorm,
+        use_bfloat16=use_bfloat16,
         mode=mode)
     decoder_blocks.append(decoder_block)
+    if half_before_layer and layer_idx == half_before_layer - 1:
+      decoder_blocks.append(tl.ReversibleConcatenatePair())
+    if double_after_layer and layer_idx == double_after_layer:
+      decoder_blocks.append(tl.ReversibleConcatenatePair())
+
+  dense_loss_layer = tl.SparseDenseWithOptions(
+      output_vocab_size,
+      d_input=d_model2,
+      sparsity_type=loss_sparsity_type,
+      sparsity=loss_sparsity,
+      d_lowrank=loss_d_lowrank,
+      prob_sparse=loss_sparsity_prob,
+      use_bfloat16=use_bfloat16,
+      mode=mode)
+
+  # Layers to merge encoder and decoder, see below for details.
+  if reversible_encoder:
+    encdec_layers = [
+        tl.ReversibleSelect([0, 1, 4, 2, 3]),  # vec_e vec_d mask_e tok_e tok_d
+        t2.ConcatWithPadding2(mode=mode),      # vec_ed vec_ed tok_e tok_d
+    ]
+  else:
+    encdec_layers = [
+        tl.ReversibleSelect([0, 3, 1, 2]),     # vec_e vec_d mask_e tok_e tok_d
+        t2.ConcatWithPadding(mode=mode),       # vec_ed tok_e tok_d
+        tl.ReversibleSelect([0, 0]),           # vec_ed vec_ed tok_e tok_d
+    ]
 
   # Assemble and return the model.
   return tl.Serial(
@@ -688,6 +821,12 @@ def Reformer2(input_vocab_size,
       tl.Parallel(in_encoder, [], [],            # vec_e tok_e tok_e vec_d tok_d
                   [tl.ShiftRight(mode=mode), out_encoder]),
 
+      # Predict mode doesn't work with padding in encoder. Raising an exception
+      # in jitted function isn't possible, so the second next best thing is
+      # to convert every embedding to NaNs, so the user will not get subtly
+      # wrong results, but clearly wrong results.
+      (_ConvertToNaNsOnAnyZero() if mode == 'predict' else []),
+
       tl.Parallel([], [tl.PaddingMask(),
                        tl.Fn('Squeeze',
                              lambda x: jnp.squeeze(x, (1, 2)), n_out=1)]),
@@ -696,17 +835,12 @@ def Reformer2(input_vocab_size,
       # Encode.
       encoder,                                  # vec_e mask_e tok_e vec_d tok_d
 
-      # Decode.
-      tl.Select([3, 0, 1, 2]),                 #  vec_d vec_e mask_e tok_e tok_d
-
       # Concat encoder and decoder, given encoder mask.
-      tl.Select([1, 0]),                       # vec_e vec_d mask_e tok_e tok_d
-      t2.ConcatWithPadding(mode=mode),         # vec_ed tok_e tok_d
+      encdec_layers,
 
-      # Run (encoder and) decoder blocks.
-      tl.Dup(),                                    # vec_ed1 vec_ed2 tok_e tok_d
-      _ReversibleSerialForget(decoder_blocks, d_model,
-                              n_layers_forget),    # vec_ed1 vec_ed2 tok_e tok_d
+      # Run decoder blocks.
+      _ReversibleSerialForget(decoder_blocks, d_model2, n_layers_forget,
+                              forget_dense),    # vec_ed1 vec_ed2 tok_e tok_d
       tl.Fn('XYAvg',
             lambda x, y: (x + y) / 2.0),           # vec_ed tok_e tok_d
       tl.LayerNorm(),                              # vec_ed tok_e tok_d
@@ -716,20 +850,34 @@ def Reformer2(input_vocab_size,
       t2.StripFromConcatenateWithPadding(mode=mode),  # vec_d tok_d
 
       # Map to output vocab.
-      tl.Dense(output_vocab_size),                 # vec_d tok_d
-      tl.LogSoftmax(),                             # vec_d tok_d
+      dense_loss_layer,  # vec_d tok_d
   )
 
 
-def _ReversibleSerialForget(layers, d_model, n_layers):
+def _ReversibleSerialForget(layers, d_model, n_layers, forget_dense=True):
   """ReversibleSerial but with a forgetting block every n_layers."""
   if not n_layers or len(layers) <= n_layers + 1:
     return tl.ReversibleSerial(layers)
   layers1, layers2 = layers[:n_layers], layers[n_layers:]
+
+  if forget_dense:
+    forgetting_layer = tl.Serial(
+        tl.Fn('XYAvg', lambda x, y: (x + y) / 2.0),
+        tl.Dense(d_model),
+        tl.Dup(),
+        )
+  else:
+    forgetting_layer = tl.Select([0, 1])
+
   return tl.Serial(
       tl.ReversibleSerial(layers1),
-      tl.Fn('XYAvg', lambda x, y: (x + y) / 2.0),
-      tl.Dense(d_model),
-      tl.Dup(),
-      _ReversibleSerialForget(layers2, d_model, n_layers)
+      forgetting_layer,
+      _ReversibleSerialForget(layers2, d_model, n_layers, forget_dense)
   )
+
+
+def _ConvertToNaNsOnAnyZero():
+  def _convert_to_nans(x, y):
+    # if all values in x are non-zeros, return x; otherwise return 0s
+    return jnp.where(jnp.all(y, keepdims=False), x, x/0.), y
+  return tl.Fn('ConvertToNaNsOnAnyZero', _convert_to_nans, n_out=2)
