@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,17 +19,20 @@ import gin
 
 from trax.rl import actor_critic
 from trax.rl import actor_critic_joint
+from trax.rl import envs
+from trax.rl import serialization_utils
 from trax.rl import training
 
 
 def configure_rl(*args, **kwargs):
   kwargs['module'] = 'trax.rl'
-  kwargs['blacklist'] = ['task', 'output_dir']
+  kwargs['denylist'] = ['task', 'output_dir']
   return gin.external_configurable(*args, **kwargs)
 
 
 A2C = configure_rl(actor_critic.A2C)
 AWR = configure_rl(actor_critic.AWR)
+LoopAWR = configure_rl(actor_critic.LoopAWR)
 PPO = configure_rl(actor_critic.PPO)
 SamplingAWR = configure_rl(actor_critic.SamplingAWR)
 
@@ -39,3 +42,7 @@ PPOJoint = configure_rl(actor_critic_joint.PPOJoint)
 
 PolicyGradient = configure_rl(training.PolicyGradient)
 DQN = configure_rl(training.DQN)
+
+TimeSeriesModel = gin.external_configurable(
+    serialization_utils.TimeSeriesModel, module='trax.rl'
+)
