@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -19,6 +19,7 @@
 import functools
 import os
 
+import gin
 import gym
 import numpy as np
 import tensorflow as tf
@@ -29,8 +30,12 @@ from trax import layers as tl
 from trax import shapes
 from trax import supervised
 from trax.fastmath import numpy as jnp
+from trax.optimizers import adam
 from trax.rl import advantages as rl_advantages
+from trax.rl import distributions
+from trax.rl import policy_tasks
 from trax.rl import training as rl_training
+from trax.rl import value_tasks
 from trax.supervised import lr_schedules as lr
 
 
@@ -57,10 +62,11 @@ class ActorCriticAgent(rl_training.PolicyAgent):
                n_replay_epochs=1,
                scale_value_targets=False,
                q_value=False,
-               q_value_aggregate='max',
+               q_value_aggregate='logsumexp',
                q_value_temperature=1.0,
                q_value_n_samples=1,
                q_value_normalization=False,
+               offline=False,
                **kwargs):  # Arguments of PolicyAgent come here.
     """Configures the actor-critic trainer.
 
@@ -97,6 +103,8 @@ class ActorCriticAgent(rl_training.PolicyAgent):
           baselines based on Q-values.
       q_value_normalization: How to normalize Q-values before aggregation.
           Allowed values: 'std', 'abs', `None`. If `None`, don't normalize.
+      offline: Whether to train in offline mode. This matters for some
+        algorithms, e.g. QWR.
       **kwargs: Arguments for `PolicyAgent` superclass.
     """
     self._n_shared_layers = n_shared_layers
@@ -136,6 +144,12 @@ class ActorCriticAgent(rl_training.PolicyAgent):
       if self._q_value_n_samples == self._vocab_size:
         # TODO(lukaszkaiser): set this explicitly once it's in AWR Trainer.
         self._sample_all_discrete_actions = True
+    if offline and is_discrete:
+      raise NotImplementedError(
+          'Offline training is only supported for continuous action spaces for '
+          'now.'
+      )
+    self._offline = offline
 
     if q_value:
       value_model = functools.partial(value_model,
@@ -145,7 +159,8 @@ class ActorCriticAgent(rl_training.PolicyAgent):
     self._value_eval_model = value_model(mode='eval')
     self._value_eval_model.init(self._value_model_signature)
     self._value_eval_jit = tl.jit_forward(
-        self._value_eval_model.pure_fn, fastmath.device_count(), do_mean=False)
+        self._value_eval_model.pure_fn, fastmath.local_device_count(),
+        do_mean=False)
 
     # Initialize policy training.
     super().__init__(task, **kwargs)
@@ -231,7 +246,7 @@ class ActorCriticAgent(rl_training.PolicyAgent):
       log_probs = None
       inputs = (observations,)
 
-    n_devices = fastmath.device_count()
+    n_devices = fastmath.local_device_count()
     weights = tl.for_n_devices(self._value_eval_model.weights, n_devices)
     state = tl.for_n_devices(self._value_eval_model.state, n_devices)
     rng = self._value_eval_model.rng
@@ -280,6 +295,12 @@ class ActorCriticAgent(rl_training.PolicyAgent):
     values *= scale
     return np.array(values)  # Move the values to CPU.
 
+  def _get_dist_inputs(self, trajectory):
+    if not self._offline:
+      return trajectory.dist_inputs
+    else:
+      return trajectory.actions
+
   def value_batches_stream(self):
     """Use the RLTask self._task to create inputs to the value model."""
     max_slice_length = self._max_slice_length + self._added_policy_slice_length
@@ -290,8 +311,9 @@ class ActorCriticAgent(rl_training.PolicyAgent):
         margin=self._added_policy_slice_length,
         epochs=self._replay_epochs,
     ):
+      dist_inputs = self._get_dist_inputs(np_trajectory)
       (values, _, act_log_probs) = self._run_value_model(
-          np_trajectory.observations, np_trajectory.dist_inputs
+          np_trajectory.observations, dist_inputs
       )
       values = self._aggregate_values(
           values, self._q_value_aggregate, act_log_probs)
@@ -348,9 +370,10 @@ class ActorCriticAgent(rl_training.PolicyAgent):
         epochs=self._replay_epochs,
         max_slice_length=max_slice_length,
         margin=self._added_policy_slice_length,
-        include_final_state=False):
+    ):
+      dist_inputs = self._get_dist_inputs(np_trajectory)
       (values, _, act_log_probs) = self._run_value_model(
-          np_trajectory.observations, np_trajectory.dist_inputs)
+          np_trajectory.observations, dist_inputs)
       values = self._aggregate_values(values, 'mean', act_log_probs)
       if len(values.shape) != 2:
         raise ValueError('Values are expected to have shape ' +
@@ -452,9 +475,6 @@ def _copy_model_weights_and_state(  # pylint: disable=invalid-name
     # pylint: enable=protected-access
 
 
-### Implementations of common actor-critic algorithms.
-
-
 class AdvantageBasedActorCriticAgent(ActorCriticAgent):
   """Base class for advantage-based actor-critic algorithms."""
 
@@ -464,6 +484,7 @@ class AdvantageBasedActorCriticAgent(ActorCriticAgent):
       advantage_estimator=rl_advantages.td_lambda,
       advantage_normalization=True,
       advantage_normalization_epsilon=1e-5,
+      advantage_normalization_factor=1.0,
       added_policy_slice_length=0,
       **kwargs
   ):
@@ -472,6 +493,7 @@ class AdvantageBasedActorCriticAgent(ActorCriticAgent):
     )
     self._advantage_normalization = advantage_normalization
     self._advantage_normalization_epsilon = advantage_normalization_epsilon
+    self._advantage_normalization_factor = advantage_normalization_factor
     super().__init__(
         task, added_policy_slice_length=added_policy_slice_length, **kwargs
     )
@@ -492,7 +514,8 @@ class AdvantageBasedActorCriticAgent(ActorCriticAgent):
     act = trajectory.actions[:, :advantages.shape[1]]
     mask = trajectory.mask[:, :advantages.shape[1]]  # Mask to zero-out padding.
     if trajectory.dist_inputs is not None:
-      dist_inputs = trajectory.dist_inputs[:, :advantages.shape[1]]
+      dist_inputs = self._get_dist_inputs(trajectory)
+      dist_inputs = dist_inputs[:, :advantages.shape[1]]
     else:
       dist_inputs = jnp.zeros(advantages.shape + (self._policy_dist.n_inputs,))
     # Shape checks to help debugging.
@@ -523,7 +546,7 @@ class AdvantageBasedActorCriticAgent(ActorCriticAgent):
 
   def _preprocess_advantages(self, advantages):
     if self._advantage_normalization:
-      advantages = (
+      advantages = self._advantage_normalization_factor * (
           (advantages - jnp.mean(advantages)) /
           (jnp.std(advantages) + self._advantage_normalization_epsilon)
       )
@@ -572,6 +595,282 @@ class AdvantageBasedActorCriticAgent(ActorCriticAgent):
     ])
 
 
+# TODO(pkozakowski): Move to a better place.
+@gin.configurable(module='trax.rl')
+def every(n_steps):
+  """Returns True every n_steps, for use as *_at functions in various places."""
+  return lambda step: step % n_steps == 0
+
+
+# TODO(pkozakowski): Rewrite all interleaved actor-critic algos to subclass
+# this, then rename to ActorCriticAgent and remove the other base classes.
+class LoopActorCriticAgent(rl_training.Agent):
+  """Base class for actor-critic algorithms based on `Loop`."""
+
+  on_policy = None
+
+  def __init__(
+      self, task, model_fn,
+      optimizer=adam.Adam,
+      policy_lr_schedule=lr.multifactor,
+      policy_n_steps_per_epoch=1000,
+      policy_weight_fn=(lambda x: x),
+      value_lr_schedule=lr.multifactor,
+      value_n_steps_per_epoch=1000,
+      value_sync_at=(lambda x: x % 100 == 0),
+      advantage_estimator=rl_advantages.monte_carlo,
+      batch_size=64,
+      network_eval_at=None,
+      n_eval_batches=1,
+      max_slice_length=1,
+      margin=0,
+      n_replay_epochs=1,
+      **kwargs
+  ):
+    """Initializes LoopActorCriticAgent.
+
+    Args:
+      task: `RLTask` instance to use.
+      model_fn: Function mode -> Trax model, building a joint policy and value
+        network.
+      optimizer: Optimizer for the policy and value networks.
+      policy_lr_schedule: Learning rate schedule for the policy network.
+      policy_n_steps_per_epoch: Number of steps to train the policy network for
+        in each epoch.
+      policy_weight_fn: Function advantages -> weights for calculating the
+        log probability weights in policy training.
+      value_lr_schedule: Learning rate schedule for the value network.
+      value_n_steps_per_epoch: Number of steps to train the value network for
+        in each epoch.
+      value_sync_at: Function step -> bool indicating when to synchronize the
+        target network with the trained network in value training.
+      advantage_estimator: Advantage estimator to use in policy and value
+        training.
+      batch_size: Batch size for training the networks.
+      network_eval_at: Function step -> bool indicating in when to evaluate the
+        networks.
+      n_eval_batches: Number of batches to compute the network evaluation
+        metrics on.
+      max_slice_length: Maximum length of a trajectory slice to train on.
+      margin: Number of timesteps to add at the end of each trajectory slice for
+        better advantage estimation.
+      n_replay_epochs: Number of epochs of trajectories to store in the replay
+        buffer.
+      **kwargs: Keyword arguments forwarded to Agent.
+    """
+    super().__init__(task, **kwargs)
+
+    self._policy_dist = distributions.create_distribution(
+        self.task.action_space
+    )
+    model_fn = functools.partial(
+        model_fn,
+        policy_distribution=self._policy_dist,
+    )
+    train_model = model_fn(mode='train')
+    eval_model = model_fn(mode='eval')
+
+    trajectory_batch_stream = self._init_trajectory_batch_stream(
+        batch_size, max_slice_length, margin, n_replay_epochs
+    )
+    advantage_estimator = advantage_estimator(task.gamma, margin=margin)
+    (value_train_task, value_eval_task) = self._init_value_tasks(
+        trajectory_batch_stream,
+        optimizer=optimizer(),
+        lr_schedule=value_lr_schedule(),
+        advantage_estimator=advantage_estimator,
+        train_model=train_model,
+        eval_model=eval_model,
+        sync_at=value_sync_at,
+        n_steps_per_epoch=value_n_steps_per_epoch,
+        n_eval_batches=n_eval_batches,
+    )
+    (policy_train_task, policy_eval_task) = self._init_policy_tasks(
+        trajectory_batch_stream,
+        optimizer=optimizer(),
+        lr_schedule=policy_lr_schedule(),
+        advantage_estimator=advantage_estimator,
+        value_train_task=value_train_task,
+        weight_fn=policy_weight_fn,
+        n_eval_batches=n_eval_batches,
+    )
+    self._init_loop(
+        train_model=train_model,
+        eval_model=eval_model,
+        policy_train_and_eval_task=(policy_train_task, policy_eval_task),
+        value_train_and_eval_task=(value_train_task, value_eval_task),
+        eval_at=network_eval_at,
+        policy_n_steps_per_epoch=policy_n_steps_per_epoch,
+        value_n_steps_per_epoch=value_n_steps_per_epoch,
+    )
+    self._init_collection(model_fn, policy_train_task.sample_batch)
+
+  def _init_trajectory_batch_stream(
+      self, batch_size, max_slice_length, margin, n_replay_epochs
+  ):
+    assert self.on_policy is not None, 'Attribute "on_policy" not set.'
+    if self.on_policy:
+      assert n_replay_epochs == 1, (
+          'Non-unit replay buffer size only makes sense for off-policy '
+          'algorithms.'
+      )
+    self._task.set_n_replay_epochs(n_replay_epochs)
+    self._max_slice_length = max_slice_length
+    return self._task.trajectory_batch_stream(
+        batch_size,
+        epochs=[-(ep + 1) for ep in range(n_replay_epochs)],
+        min_slice_length=(1 + margin),
+        max_slice_length=(self._max_slice_length + margin),
+        margin=margin,
+    )
+
+  def _init_value_tasks(
+      self,
+      trajectory_batch_stream,
+      optimizer,
+      lr_schedule,
+      advantage_estimator,
+      train_model,
+      eval_model,
+      sync_at,
+      n_steps_per_epoch,
+      n_eval_batches,
+  ):
+    def sync_also_at_epoch_boundaries(step):
+      return sync_at(step) or (
+          # 0 - end of the epoch, 1 - beginning of the next.
+          step % n_steps_per_epoch in (0, 1)
+      )
+
+    head_selector = tl.Select([1])
+    value_train_task = value_tasks.ValueTrainTask(
+        trajectory_batch_stream,
+        optimizer,
+        lr_schedule,
+        advantage_estimator=advantage_estimator,
+        model=train_model,
+        target_model=eval_model,
+        target_scale=(1 - self.task.gamma),
+        sync_at=sync_also_at_epoch_boundaries,
+        head_selector=head_selector,
+    )
+    value_eval_task = value_tasks.ValueEvalTask(
+        value_train_task, n_eval_batches, head_selector
+    )
+    return (value_train_task, value_eval_task)
+
+  def _init_policy_tasks(
+      self,
+      trajectory_batch_stream,
+      optimizer,
+      lr_schedule,
+      advantage_estimator,
+      value_train_task,
+      weight_fn,
+      n_eval_batches,
+  ):
+    head_selector = tl.Select([0], n_in=2)
+    policy_train_task = policy_tasks.PolicyTrainTask(
+        trajectory_batch_stream,
+        optimizer,
+        lr_schedule,
+        self._policy_dist,
+        advantage_estimator=advantage_estimator,
+        value_fn=value_train_task.value,
+        weight_fn=weight_fn,
+        head_selector=head_selector,
+    )
+    policy_eval_task = policy_tasks.PolicyEvalTask(
+        policy_train_task, n_eval_batches, head_selector
+    )
+    return (policy_train_task, policy_eval_task)
+
+  def _init_loop(
+      self,
+      train_model,
+      eval_model,
+      policy_train_and_eval_task,
+      value_train_and_eval_task,
+      eval_at,
+      policy_n_steps_per_epoch,
+      value_n_steps_per_epoch,
+  ):
+    (policy_train_task, policy_eval_task) = policy_train_and_eval_task
+    (value_train_task, value_eval_task) = value_train_and_eval_task
+
+    if self._output_dir is not None:
+      model_output_dir = os.path.join(self._output_dir, 'model')
+    else:
+      model_output_dir = None
+
+    self._n_train_steps_per_epoch = (
+        policy_n_steps_per_epoch + value_n_steps_per_epoch
+    )
+
+    checkpoint_at = lambda step: step % self._n_train_steps_per_epoch == 0
+
+    def which_task(step):
+      if step % self._n_train_steps_per_epoch < value_n_steps_per_epoch:
+        return 1
+      else:
+        return 0
+
+    self._loop = supervised.training.Loop(
+        model=train_model,
+        tasks=(policy_train_task, value_train_task),
+        eval_model=eval_model,
+        eval_tasks=(policy_eval_task, value_eval_task),
+        output_dir=model_output_dir,
+        eval_at=eval_at,
+        checkpoint_at=checkpoint_at,
+        which_task=which_task,
+    )
+
+    # Validate the restored checkpoints.
+    # TODO(pkozakowski): Move this to the base class once all Agents use Loop.
+    if self._loop.step != self._epoch * self._n_train_steps_per_epoch:
+      raise ValueError(
+          'The number of Loop steps must equal the number of Agent epochs '
+          'times the number of steps per epoch, got {}, {} and {}.'.format(
+              self._loop.step, self._epoch, self._n_train_steps_per_epoch
+          )
+      )
+
+  def _init_collection(self, model_fn, sample_batch):
+    self._collect_model = model_fn(mode='collect')
+    self._collect_model.init(shapes.signature(sample_batch))
+    if self._task._initial_trajectories == 0:  # pylint: disable=protected-access
+      self._collect_trajectories()
+
+  @property
+  def loop(self):
+    """Loop exposed for testing."""
+    return self._loop
+
+  def policy(self, trajectory, temperature=1.0):
+    """Policy function that allows to play using this agent."""
+    tr_slice = trajectory[-self._max_slice_length:]
+    trajectory_np = tr_slice.to_np(timestep_to_np=self.task.timestep_to_np)
+    return rl_training.network_policy(
+        collect_model=self._collect_model,
+        policy_distribution=self._policy_dist,
+        loop=self.loop,
+        trajectory_np=trajectory_np,
+        head_index=0,
+        temperature=temperature,
+    )
+
+  def train_epoch(self):
+    """Trains RL for one epoch."""
+    # Copy policy state accumulated during data collection to the trainer.
+    self._loop.update_weights_and_state(state=self._collect_model.state)
+    # Perform one gradient step per training epoch to ensure we stay on policy.
+    self._loop.run(n_steps=self._n_train_steps_per_epoch)
+
+
+### Implementations of common actor-critic algorithms.
+
+
 class A2C(AdvantageBasedActorCriticAgent):
   """Trains policy and value models using the A2C algorithm."""
 
@@ -684,15 +983,44 @@ class PPO(AdvantageBasedActorCriticAgent):
     return tl.Fn('PPOLoss', f)
 
 
+def _weighted_percentiles(x, thresholds):
+  """Calculate weights for x by percentile-and-weights given in thresholds.
+
+  Thresholds contain a list of (p, weight, minumum). For each threshold,
+  all elements of x that are above the p-th percentile *and* above minimum
+  get the weight weight, and all other get the weight 0.
+  The result is the sum over all thresholds.
+
+  Args:
+    x: tensor to calculate the weights for
+    thresholds: list of triples (percentile, weight, minimum) used to
+      calculate the weights (see above how)
+
+  Returns:
+    weights, a tensor of the same shape as x
+  """
+  res = []
+  for (percentile, weight, minimum) in thresholds:
+    threshold = jnp.percentile(x, percentile)
+    if minimum is not None:
+      threshold = jnp.maximum(minimum, threshold)
+    zero_ones = jnp.where(x < threshold, jnp.zeros_like(x), jnp.ones_like(x))
+    res.append(weight * zero_ones)
+  return sum(res)
+
+
 # AWR is an off-policy actor-critic RL algorithm.
-def awr_weights(advantages, beta):
+def awr_weights(advantages, beta, thresholds):
+  if thresholds:
+    return _weighted_percentiles(advantages, thresholds)
   return jnp.exp(advantages / beta)
 
 
 # Helper functions for computing AWR metrics.
-def awr_metrics(beta, preprocess_layer=None):
+def awr_metrics(beta, thresholds, preprocess_layer=None):
   return {  # pylint: disable=g-complex-comprehension
-      'awr_weight_' + name: awr_weight_stat(name, fn, beta, preprocess_layer)
+      'awr_weight_' + name: awr_weight_stat(name, fn, beta, thresholds,
+                                            preprocess_layer)
       for (name, fn) in [
           ('mean', jnp.mean),
           ('std', jnp.std),
@@ -702,23 +1030,23 @@ def awr_metrics(beta, preprocess_layer=None):
   }
 
 
-def awr_weight_stat(stat_name, stat_fn, beta, preprocess_layer):
+def awr_weight_stat(stat_name, stat_fn, beta, thresholds, preprocess_layer):
   # Select just the advantages if preprocess layer is not given.
   preprocess = tl.Select([1]) if preprocess_layer is None else preprocess_layer
   return tl.Serial([
       preprocess,
       tl.Fn(
           'AWRWeight' + stat_name.capitalize(),
-          lambda x: stat_fn(awr_weights(x, beta)),
+          lambda x: stat_fn(awr_weights(x, beta, thresholds)),
       ),
   ])
 
 
-def AWRLoss(beta, w_max):  # pylint: disable=invalid-name
+def AWRLoss(beta, w_max, thresholds):  # pylint: disable=invalid-name
   """Definition of the Advantage Weighted Regression (AWR) loss."""
   def f(log_probs, advantages, old_log_probs, mask):
     del old_log_probs  # Not used in AWR.
-    weights = jnp.minimum(awr_weights(advantages, beta), w_max)
+    weights = jnp.minimum(awr_weights(advantages, beta, thresholds), w_max)
     return -jnp.sum(log_probs * weights * mask) / jnp.sum(mask)
   return tl.Fn('AWRLoss', f)
 
@@ -728,32 +1056,42 @@ class AWR(AdvantageBasedActorCriticAgent):
 
   on_policy = False
 
-  def __init__(self, task, beta=1.0, w_max=20.0, **kwargs):
+  def __init__(self, task, beta=1.0, w_max=20.0, thresholds=None, **kwargs):
     """Configures the AWR Trainer."""
     self._beta = beta
     self._w_max = w_max
+    self._thresholds = thresholds
     super().__init__(task, **kwargs)
 
   @property
   def policy_loss_given_log_probs(self):
     """Policy loss."""
-    return AWRLoss(beta=self._beta, w_max=self._w_max)  # pylint: disable=no-value-for-parameter
-
-  @property
-  def policy_metrics(self):
-    metrics = super().policy_metrics
-    metrics.update(awr_metrics(self._beta))
-    return metrics
+    return AWRLoss(beta=self._beta, w_max=self._w_max,
+                   thresholds=self._thresholds)  # pylint: disable=no-value-for-parameter
 
 
-def SamplingAWRLoss(beta, w_max, reweight=False, sampled_all_discrete=False):  # pylint: disable=invalid-name
+class LoopAWR(LoopActorCriticAgent):
+  """Advantage Weighted Regression."""
+
+  on_policy = False
+
+  def __init__(self, task, model_fn, beta=1.0, w_max=20, **kwargs):
+    def policy_weight_fn(advantages):
+      return jnp.minimum(jnp.exp(advantages / beta), w_max)
+    super().__init__(
+        task, model_fn, policy_weight_fn=policy_weight_fn, **kwargs
+    )
+
+
+def SamplingAWRLoss(beta, w_max, thresholds,  # pylint: disable=invalid-name
+                    reweight=False, sampled_all_discrete=False):
   """Definition of the Advantage Weighted Regression (AWR) loss."""
   def f(log_probs, advantages, old_log_probs, mask):
     if reweight:  # Use new policy weights for sampled actions instead.
       mask *= jnp.exp(fastmath.stop_gradient(log_probs) - old_log_probs)
     if sampled_all_discrete:  # Actions were sampled uniformly; weight them.
       mask *= jnp.exp(old_log_probs)
-    weights = jnp.minimum(awr_weights(advantages, beta), w_max)
+    weights = jnp.minimum(awr_weights(advantages, beta, thresholds), w_max)
     return -jnp.sum(log_probs * weights * mask) / jnp.sum(mask)
   return tl.Fn('SamplingAWRLoss', f)
 
@@ -763,10 +1101,12 @@ class SamplingAWR(AdvantageBasedActorCriticAgent):
 
   on_policy = False
 
-  def __init__(self, task, beta=1.0, w_max=20.0, reweight=False, **kwargs):
+  def __init__(self, task, beta=1.0, w_max=20.0, thresholds=None,
+               reweight=False, **kwargs):
     """Configures the AWR Trainer."""
     self._beta = beta
     self._w_max = w_max
+    self._thresholds = thresholds
     self._reweight = reweight
     super().__init__(task, q_value=True, **kwargs)
 
@@ -800,7 +1140,8 @@ class SamplingAWR(AdvantageBasedActorCriticAgent):
         )
     }
     metrics.update(awr_metrics(
-        self._beta, preprocess_layer=self._policy_inputs_to_advantages(True)))
+        self._beta, self._thresholds,
+        preprocess_layer=self._policy_inputs_to_advantages(True)))
     return metrics
 
   @property
@@ -833,7 +1174,8 @@ class SamplingAWR(AdvantageBasedActorCriticAgent):
         # Policy loss is expected to consume
         # (log_probs, advantages, old_log_probs, mask).
         SamplingAWRLoss(
-            beta=self._beta, w_max=self._w_max, reweight=self._reweight,
+            beta=self._beta, w_max=self._w_max, thresholds=self._thresholds,
+            reweight=self._reweight,
             sampled_all_discrete=self._sample_all_discrete_actions)
     )
 
@@ -844,10 +1186,10 @@ class SamplingAWR(AdvantageBasedActorCriticAgent):
         self._policy_batch_size,
         epochs=self._replay_epochs,
         max_slice_length=self._max_slice_length,
-        include_final_state=False,
     ):
+      dist_inputs = self._get_dist_inputs(np_trajectory)
       (q_values, actions, act_log_probs) = self._run_value_model(
-          np_trajectory.observations, np_trajectory.dist_inputs)
+          np_trajectory.observations, dist_inputs)
       shapes.assert_same_shape(q_values, act_log_probs)
 
       # q_values shape: (batch_size, n_samples, length)
