@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -322,10 +322,10 @@ class PPOJoint(ActorCriticJointAgent):
           f'values.shape was {values.shape}')
 
       entropy_loss = rl_layers.EntropyLoss(
-          dist_inputs, actions,
-          log_prob_fun=self._policy_dist.log_prob,
-          entropy_coeff=self._entropy_coeff,
-          entropy_fun=self._policy_dist.entropy)
+          dist_inputs,
+          distribution=self._policy_dist,
+          coeff=self._entropy_coeff,
+      )
 
       assert jnp.ndim(entropy_loss) == 0, f'entropy_loss was {entropy_loss}'
 
@@ -374,11 +374,12 @@ class PPOJoint(ActorCriticJointAgent):
   def entropy_loss(self):
     """Entropy layer."""
     def f(dist_inputs, values, returns, dones, rewards, actions):
-      del values, returns, dones, rewards
+      del values, returns, dones, rewards, actions
       return rl_layers.EntropyLoss(
-          dist_inputs, actions, log_prob_fun=self._policy_dist.log_prob,
-          entropy_coeff=self._entropy_coeff,
-          entropy_fun=self._policy_dist.entropy)
+          dist_inputs,
+          distribution=self._policy_dist,
+          coeff=self._entropy_coeff,
+      )
     return tl.Fn('EntropyLoss', f)
 
   @property
@@ -542,10 +543,10 @@ class A2CJoint(ActorCriticJointAgent):
       assert jnp.ndim(a2c_objective) == 0, f'a2c_objective was {a2c_objective}'
 
       entropy_loss = rl_layers.EntropyLoss(
-          dist_inputs, actions,
-          log_prob_fun=self._policy_dist.log_prob,
-          entropy_coeff=self._entropy_coeff,
-          entropy_fun=self._policy_dist.entropy)
+          dist_inputs,
+          distribution=self._policy_dist,
+          coeff=self._entropy_coeff,
+      )
 
       assert jnp.ndim(entropy_loss) == 0, f'entropy_loss was {entropy_loss}'
 
@@ -564,11 +565,12 @@ class A2CJoint(ActorCriticJointAgent):
   def entropy_loss(self):
     """Entropy layer."""
     def f(dist_inputs, values, returns, dones, rewards, actions):
-      del values, returns, dones, rewards
+      del values, returns, dones, rewards, actions
       return rl_layers.EntropyLoss(
-          dist_inputs, actions, log_prob_fun=self._policy_dist.log_prob,
-          entropy_coeff=self._entropy_coeff,
-          entropy_fun=self._policy_dist.entropy)
+          dist_inputs,
+          distribution=self._policy_dist,
+          coeff=self._entropy_coeff,
+      )
     return tl.Fn('EntropyLoss', f)
 
   @property
@@ -623,10 +625,11 @@ class AWRJoint(ActorCriticJointAgent):
 
   # TODO(henrykm): value_loss_coeff looks like a common parameter
   def __init__(self, task, value_loss_coeff=0.1, beta=1.0, w_max=20.0,
-               **kwargs):
+               thresholds=None, **kwargs):
     """Configures the joint AWR Trainer."""
     self._beta = beta
     self._w_max = w_max
+    self._thresholds = thresholds
     self._value_loss_coeff = value_loss_coeff
     super().__init__(task, **kwargs)
 
@@ -648,8 +651,9 @@ class AWRJoint(ActorCriticJointAgent):
     def f(preds, values, returns, actions, mask):
       advantages = jnp.squeeze(returns - stop_gradient(values), axis=-1)
       logps = self._policy_dist.log_prob(preds, actions)
-      awr_loss = actor_critic.AWRLoss(beta=self._beta, w_max=self._w_max)(
-          (logps, advantages, jnp.zeros_like(logps), mask))
+      awr_loss = actor_critic.AWRLoss(
+          beta=self._beta, w_max=self._w_max, thresholds=self._thresholds)(
+              (logps, advantages, jnp.zeros_like(logps), mask))
       l2_value_loss = jnp.mean((returns - values)**2) * self._value_loss_coeff
       return awr_loss + l2_value_loss
     return tl.Fn('AWRJointLoss', f)
