@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -69,7 +69,7 @@ class Distribution:
                  lambda inputs, point: self.log_prob(inputs, point))  # pylint: disable=unnecessary-lambda
 
 
-@gin.configurable(blacklist=['n_categories', 'shape'])
+@gin.configurable(denylist=['n_categories', 'shape'])
 class Categorical(Distribution):
   """Categorical distribution parametrized by logits."""
 
@@ -115,7 +115,7 @@ class Categorical(Distribution):
     return -jnp.sum(probs * log_probs, axis=-1)
 
 
-@gin.configurable(blacklist=['shape'])
+@gin.configurable(denylist=['shape'])
 class Gaussian(Distribution):
   """Independent multivariate Gaussian distribution parametrized by mean."""
 
@@ -139,7 +139,13 @@ class Gaussian(Distribution):
 
   def _params(self, inputs):
     """Extracts the mean and std parameters from the inputs."""
-    assert inputs.shape[-1] == self.n_inputs
+    if inputs.shape[-1] != self.n_inputs:
+      raise ValueError(
+          'Invalid distribution parametrization - expected {} parameters, '
+          'got {}. Input shape: {}.'.format(
+              self.n_inputs, inputs.shape[-1], inputs.shape
+          )
+      )
     n_dims = self._n_dims
     # Split the distribution inputs into two parts: mean and std.
     mean = inputs[..., :n_dims]
