@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -27,7 +27,8 @@ from trax import fastmath
 from trax.supervised import training
 
 
-class _TimeStep(object):
+
+class _TimeStep:
   """A single step of interaction with a RL environment.
 
   TimeStep stores a single step in the trajectory of an RL run:
@@ -79,10 +80,10 @@ TimeStepNp = collections.namedtuple('TimeStepNp', [
 ])
 
 
-class Trajectory(object):
+class Trajectory:
   """A trajectory of interactions with a RL environment.
 
-  Trajectories are created when interacting with a RL environment. They can
+  Trajectories are created when interacting with an RL environment. They can
   be prolonged and sliced and when completed, allow to re-calculate returns.
   """
 
@@ -91,16 +92,18 @@ class Trajectory(object):
     # reuse code from base_trainer.dump_trajectories and related functions.
     if observation is not None:
       self._timesteps = [_TimeStep(observation)]
+    self._trajectory_np = None
+    self._cached_to_np_args = None
 
   def __len__(self):
     return len(self._timesteps)
 
   def __str__(self):
-    return str([(ts.observation, ts.action, ts.reward)
+    return str([(ts.observation, ts.action, ts.reward, ts.done)
                 for ts in self._timesteps])
 
   def __repr__(self):
-    return repr([(ts.observation, ts.action, ts.reward)
+    return repr([(ts.observation, ts.action, ts.reward, ts.done)
                  for ts in self._timesteps])
 
   def __getitem__(self, key):
@@ -123,7 +126,23 @@ class Trajectory(object):
     last_timestep = self._timesteps[-1]
     return last_timestep.observation
 
-  def extend(self, action, dist_inputs, reward, done, new_observation, mask=1):
+  @property
+  def done(self):
+    """Returns whether the trajectory is finished."""
+    if len(self._timesteps) < 2:
+      return False
+    second_last_timestep = self._timesteps[-2]
+    return second_last_timestep.done
+
+  @done.setter
+  def done(self, done):
+    """Sets the `done` flag in the last timestep."""
+    if len(self._timesteps) < 2:
+      raise ValueError('No interactions yet in the trajectory.')
+    last_timestep = self._timesteps[-2]
+    last_timestep.done = done
+
+  def extend(self, action, dist_inputs, new_observation, reward, done, mask=1):
     """Take action in the last state, getting reward and going to new state."""
     last_timestep = self._timesteps[-1]
     last_timestep.action = action
@@ -154,12 +173,22 @@ class Trajectory(object):
         mask=ts.mask,
     ))
 
-  def to_np(self, timestep_to_np=None):
+  def to_np(self, margin=0, timestep_to_np=None):
     """Create a tuple of numpy arrays from a given trajectory."""
+    timestep_to_np = timestep_to_np or self._default_timestep_to_np
+    args = (margin, timestep_to_np)
+
+    # Return the cached result if the arguments agree and the trajectory has not
+    # grown.
+    if self._trajectory_np:
+      result_length = len(self) + margin - 1
+      length_ok = self._trajectory_np.observations.shape[0] == result_length
+      if args == self._cached_to_np_args and length_ok:
+        return self._trajectory_np
+
     observations, actions, dist_inputs, rewards, returns, dones, masks = (
         [], [], [], [], [], [], []
     )
-    timestep_to_np = timestep_to_np or self._default_timestep_to_np
     for timestep in self._timesteps:
       if timestep.action is None:
         obs = timestep_to_np(timestep).observation
@@ -174,12 +203,32 @@ class Trajectory(object):
         returns.append(timestep_np.return_)
         masks.append(timestep_np.mask)
 
+    # TODO(pkozakowski): The case len(obs) == 1 is for handling
+    # "dummy trajectories", that are only there to determine data shapes. Check
+    # if they're still required.
+    if len(observations) > 1:
+      # Extend the trajectory with a given margin - this is to make sure that
+      # the networks always "see" the "done" states in the training data, even
+      # when a suffix is added to the trajectory slice for better estimation of
+      # returns.
+      # We set `mask` to 0, so the added timesteps don't influence the loss. We
+      # set `done` to True for easier implementation of advantage estimators.
+      # The rest of the fields don't matter, so we set them to 0 for easy
+      # debugging (unless they're None). The list of observations is longer, so
+      # we pad it with margin - 1.
+      masks.extend([0] * margin)
+      dones.extend([True] * margin)
+      observations.extend([np.zeros_like(observations[-1])] * (margin - 1))
+      for x in (actions, dist_inputs, rewards, returns):
+        filler = None if x[-1] is None else np.zeros_like(x[-1])
+        x.extend([filler] * margin)
+
     def stack(x):
       if not x:
         return None
       return fastmath.nested_stack(x)
 
-    return TrajectoryNp(**{  # pylint: disable=g-complex-comprehension
+    trajectory_np = TrajectoryNp(**{  # pylint: disable=g-complex-comprehension
         key: stack(value) for (key, value) in [
             ('observations', observations),
             ('actions', actions),
@@ -191,8 +240,13 @@ class Trajectory(object):
         ]
     })
 
+    self._trajectory_np = trajectory_np
+    self._cached_to_np_args = args
 
-def play(env, policy, dm_suite=False, max_steps=None):
+    return trajectory_np
+
+
+def play(env, policy, dm_suite=False, max_steps=None, last_observation=None):
   """Play an episode in env taking actions according to the given policy.
 
   Environment is first reset and an from then on, a game proceeds. At each
@@ -208,30 +262,33 @@ def play(env, policy, dm_suite=False, max_steps=None):
       defined as the log of the probability of taking that action).
     dm_suite: whether we are using the DeepMind suite or the gym interface
     max_steps: for how many steps to play.
+    last_observation: last observation from a previous trajectory slice, used to
+      begin a new one. Controls whether we reset the environment at the
+      beginning - if `None`, resets the env and starts the slice from the
+      observation got from reset().
 
   Returns:
-    a completed trajectory that was just played.
+    a completed trajectory slice that was just played.
   """
-  terminal = False
+  done = False
   cur_step = 0
-  if dm_suite:
-    cur_trajectory = Trajectory(env.reset().observation)
-    while not terminal and (max_steps is None or cur_step < max_steps):
-      action, dist_inputs = policy(cur_trajectory)
-      observation = env.step(action)
-      cur_trajectory.extend(action, dist_inputs,
-                            observation.reward,
-                            observation.step_type.last(),
-                            observation.observation)
-      cur_step += 1
-      terminal = observation.step_type.last()
-  else:
-    cur_trajectory = Trajectory(env.reset())
-    while not terminal and (max_steps is None or cur_step < max_steps):
-      action, dist_inputs = policy(cur_trajectory)
-      observation, reward, terminal, _ = env.step(action)
-      cur_trajectory.extend(action, dist_inputs, reward, terminal, observation)
-      cur_step += 1
+  if last_observation is None:
+    # TODO(pkozakowski): Make a Gym wrapper over DM envs to get rid of branches
+    # like that.
+    last_observation = env.reset().observation if dm_suite else env.reset()
+  cur_trajectory = Trajectory(last_observation)
+  while not done and (max_steps is None or cur_step < max_steps):
+    action, dist_inputs = policy(cur_trajectory)
+    step = env.step(action)
+    if dm_suite:
+      observation_reward_done = (
+          step.observation, step.reward, step.step_type.last()
+      )
+    else:
+      observation_reward_done = step[:3]
+    cur_trajectory.extend(action, dist_inputs, *observation_reward_done)
+    cur_step += 1
+    (_, _, done) = observation_reward_done
   return cur_trajectory
 
 
@@ -252,23 +309,35 @@ def _sample_proportionally(inputs, weights):
 
   Args:
     inputs: a list, we will return one element of this list.
-    weights: a list of numbers of the same length as inputs; we will sample
+    weights: a sequence of numbers of the same length as inputs; we will sample
       the k-th input with probability weights[k] / sum(weights).
 
   Returns:
     an element from inputs.
   """
   l = len(inputs)
+  weights = np.array(weights)
   if l != len(weights):
     raise ValueError(f'Inputs and weights must have the same length, but do not'
                      f': {l} != {len(weights)}')
-  weights_sum = float(sum(weights))
-  norm_weights = [w / weights_sum for w in weights]
+  norm_weights = weights / np.sum(weights)
+  # TODO(pkozakowski): Currently this is O(n). It can be sped up to O(log n) by
+  # storing CDF and binsearching on it.
   idx = np.random.choice(l, p=norm_weights)
   return inputs[int(idx)]
 
 
-@gin.configurable()
+def _n_slices(trajectory, max_slice_length, margin):
+  """How many slices of length upto max_slice_length in a trajectory."""
+  # TODO(lukaszkaiser): add option to sample from n last trajectories.
+  if not max_slice_length:
+    return 1
+  # A trajectory [a, b, c, end_state] will have 2 slices of length 2:
+  # the slice [a, b] and the one [b, c], with margin=0; 3 with margin=1.
+  return max(1, len(trajectory) + margin - max_slice_length)
+
+
+@gin.configurable
 class RLTask:
   """A RL task: environment and a collection of trajectories."""
 
@@ -278,6 +347,7 @@ class RLTask:
                dm_suite=False,
                random_starts=True,
                max_steps=None,
+               time_limit=None,
                timestep_to_np=None,
                num_stacked_frames=1,
                n_replay_epochs=1):
@@ -294,7 +364,10 @@ class RLTask:
       gamma: float: discount factor for calculating returns.
       dm_suite: whether we are using the DeepMind suite or the gym interface
       random_starts: use random starts for training of Atari agents.
-      max_steps: Optional int: stop all trajectories at that many steps.
+      max_steps: optional int: cut all trajectory slices at that many steps.
+        The trajectory will be continued in the next epochs, up to `time_limit`.
+      time_limit: optional int: stop all trajectories after that many steps (or
+        after getting "done"). If `None`, use the same value as `max_steps`.
       timestep_to_np: a function that turns a timestep into a numpy array
         (ie., a tensor); if None, we just use the state of the timestep to
         represent it, but other representations (such as embeddings that include
@@ -305,43 +378,11 @@ class RLTask:
     if isinstance(env, str):
       self._env_name = env
       if dm_suite:
-        eval_env = environments.load_from_settings(
-            platform='atari',
-            settings={
-                'levelName': env,
-                'interleaved_pixels': True,
-                'zero_indexed_actions': True,
-                'phase': 'testing',
-            })
-        eval_env = atari_wrapper.AtariWrapper(
-            environment=eval_env,
-            max_abs_reward=None,
-            num_stacked_frames=num_stacked_frames)
-        if random_starts:
-          env = environments.load_from_settings(
-              platform='atari',
-              settings={
-                  'levelName': env,
-                  'interleaved_pixels': True,
-                  'zero_indexed_actions': True,
-                  'phase': 'training'
-              })
-        else:
-          # This is the default behavior used so far
-          # in QWR experiments (implicitly 'condition': 'at30').
-          env = environments.load_from_settings(
-              platform='atari',
-              settings={
-                  'levelName': env,
-                  'interleaved_pixels': True,
-                  'zero_indexed_actions': True
-              })
-        env = atari_wrapper.AtariWrapper(
-            environment=env,
-            num_stacked_frames=num_stacked_frames)
+        eval_env = None
+        env = None
       else:
-        env = gym.make(env)
-        eval_env = env
+        env = gym.make(self._env_name)
+        eval_env = gym.make(self._env_name)
     else:
       self._env_name = type(env).__name__
       eval_env = env
@@ -349,27 +390,32 @@ class RLTask:
     self._eval_env = eval_env
     self._dm_suite = dm_suite
     self._max_steps = max_steps
+    if time_limit is None:
+      time_limit = max_steps
+    self._time_limit = time_limit
     self._gamma = gamma
     self._initial_trajectories = initial_trajectories
+    self._last_observation = None
+    self._n_steps_left = time_limit
+    # Example trajectory for determining input/output shapes of the networks.
+    self._example_trajectory = self.play(
+        _random_policy(self.action_space), only_eval=True
+    )
     # TODO(lukaszkaiser): find a better way to pass initial trajectories,
     # whether they are an explicit list, a file, or a number of random ones.
     if isinstance(initial_trajectories, int):
-      if self._initial_trajectories > 0:
-        initial_trajectories = [
-            self.play(_random_policy(self.action_space))
-            for _ in range(initial_trajectories)
-        ]
-      else:
-        initial_trajectories = [
-            # Whatever we gather here is intended to be removed
-            # in PolicyTrainer. Here we just gather some example inputs.
-            self.play(_random_policy(self.action_space))
-        ]
+      initial_trajectories = [
+          self.play(_random_policy(self.action_space))
+          for _ in range(initial_trajectories)
+      ]
     if isinstance(initial_trajectories, str):
       initial_trajectories = self.load_initial_trajectories_from_path(
           initial_trajectories_path=initial_trajectories)
     if isinstance(initial_trajectories, list):
-      initial_trajectories = {0: initial_trajectories}
+      if initial_trajectories:
+        initial_trajectories = {0: initial_trajectories}
+      else:
+        initial_trajectories = {}
     self._timestep_to_np = timestep_to_np
     # Stored trajectories are indexed by epoch and within each epoch they
     # are stored in the order of generation so we can implement replay buffers.
@@ -498,9 +544,42 @@ class RLTask:
     if max_steps is None:
       max_steps = self._max_steps
     if only_eval:
-      cur_trajectory = play(self._eval_env, policy, self._dm_suite, max_steps)
+      cur_trajectory = play(
+          self._eval_env, policy, self._dm_suite,
+          # Only step up to the time limit.
+          max_steps=min(max_steps, self._time_limit),
+          # Always reset at the beginning of an eval episode.
+          last_observation=None,
+      )
     else:
-      cur_trajectory = play(self._env, policy, self._dm_suite, max_steps)
+      cur_trajectory = play(
+          self._env, policy, self._dm_suite,
+          # Only step up to the time limit, used up by all slices of the same
+          # trajectory.
+          max_steps=min(max_steps, self._n_steps_left),
+          # Pass the environmnent state between play() calls, so one episode can
+          # span multiple training epochs.
+          # NOTE: Cutting episodes between epochs may hurt e.g. with
+          # Transformers.
+          # TODO(pkozakowski): Join slices together if this becomes a problem.
+          last_observation=self._last_observation,
+      )
+      # Update the number of steps left to reach the time limit.
+      # NOTE: This should really be done as an env wrapper.
+      # TODO(pkozakowski): Do that once we wrap the DM envs in Gym interface.
+      # The initial reset doesn't count.
+      self._n_steps_left -= len(cur_trajectory) - 1
+      assert self._n_steps_left >= 0
+      if self._n_steps_left == 0:
+        cur_trajectory.done = True
+      # Pass the last observation between trajectory slices.
+      if cur_trajectory.done:
+        self._last_observation = None
+        # Reset the time limit.
+        self._n_steps_left = self._time_limit
+      else:
+        self._last_observation = cur_trajectory.last_observation
+
     cur_trajectory.calculate_returns(self._gamma)
     return cur_trajectory
 
@@ -514,17 +593,16 @@ class RLTask:
   ):
     """Collect experience in env playing the given policy."""
     max_steps = max_steps or self.max_steps
-    if n_trajectories:
+    if n_trajectories is not None:
       new_trajectories = [self.play(policy, max_steps=max_steps,
                                     only_eval=only_eval)
                           for _ in range(n_trajectories)]
-    elif n_interactions:
-      # TODO(pkozakowski): Test this mode of experience collection.
+    elif n_interactions is not None:
       new_trajectories = []
       while n_interactions > 0:
-        traj = self.play(policy, max_steps=max_steps)
+        traj = self.play(policy, max_steps=min(n_interactions, max_steps))
         new_trajectories.append(traj)
-        n_interactions -= len(traj)
+        n_interactions -= len(traj) - 1  # The initial reset does not count.
     else:
       raise ValueError(
           'Either n_trajectories or n_interactions must be defined.'
@@ -532,13 +610,17 @@ class RLTask:
 
     # Calculate returns.
     returns = [t.total_return for t in new_trajectories]
+    if returns:
+      mean_returns = sum(returns) / float(len(returns))
+    else:
+      mean_returns = 0
 
     # If we're only evaluating, we're done, return the average.
     if only_eval:
-      return sum(returns) / float(len(returns))
-
+      return mean_returns
     # Store new trajectories.
-    self._trajectories[epoch_id].extend(new_trajectories)
+    if new_trajectories:
+      self._trajectories[epoch_id].extend(new_trajectories)
 
     # Mark that epoch epoch_id has changed.
     if epoch_id in self._saved_epochs_unchanged:
@@ -556,7 +638,7 @@ class RLTask:
     self._n_trajectories += len(new_trajectories)
     self._n_interactions += sum([len(traj) for traj in new_trajectories])
 
-    return sum(returns) / float(len(returns))
+    return mean_returns
 
   def n_trajectories(self, epochs=None):
     # TODO(henrykm) support selection of epochs if really necessary (will
@@ -570,21 +652,32 @@ class RLTask:
     del epochs
     return self._n_interactions
 
-  def remove_epoch(self, epoch):
-    """Useful when we need to remove an unwanted trajectory."""
-    if epoch in self._trajectories:
-      self._trajectories.pop(epoch)
+  def _random_slice(self, trajectory, max_slice_length, margin):
+    """Returns a random TrajectoryNp slice from a given trajectory."""
+    # Sample a slice from the trajectory.
+    slice_start = np.random.randint(
+        _n_slices(trajectory, max_slice_length, margin)
+    )
+
+    # Convert the whole trajectory to Numpy while adding the margin. The
+    # result is cached, so we don't have to repeat this for every sample.
+    trajectory_np = trajectory.to_np(margin, self._timestep_to_np)
+
+    # Slice and yield the result.
+    slice_end = slice_start + (
+        max_slice_length or trajectory_np.observations.shape[0]
+    )
+    return fastmath.nested_map(
+        lambda x: x[slice_start:slice_end], trajectory_np
+    )
 
   def trajectory_stream(self, epochs=None, max_slice_length=None,
-                        include_final_state=False,
                         sample_trajectories_uniformly=False, margin=0):
     """Return a stream of random trajectory slices from the specified epochs.
 
     Args:
       epochs: a list of epochs to use; we use all epochs if None
       max_slice_length: maximum length of the slices of trajectories to return
-      include_final_state: whether to include slices with the final state of
-        the trajectory which may have no action and reward
       sample_trajectories_uniformly: whether to sample trajectories uniformly,
         or proportionally to the number of slices in each trajectory (default)
       margin: number of extra steps after "done" that should be included in
@@ -594,85 +687,83 @@ class RLTask:
       random trajectory slices sampled uniformly from all slices of length
       up to max_slice_length in all specified epochs
     """
-    # TODO(lukaszkaiser): add option to sample from n last trajectories.
-    end_offset = 0 if include_final_state else 1
-    def n_slices(t):
-      """How many slices of length upto max_slice_length in a trajectory."""
-      if not max_slice_length:
-        return 1
-      # A trajectory [a, b, c, end_state] will have 2 slices of length 2:
-      # the slice [a, b] and the one [b, c], with end_offset; 3 without.
-      return max(1, len(t) + margin - max_slice_length + 1 - end_offset)
+    # {int: array[int]};
+    # epoch_to_ns_slices[epoch][i] = n_slices(self._trajectories[epoch][i])
+    # It stores arrays for faster sampling.
+    epoch_to_ns_slices = {}
+    # {int: int};
+    # epoch_to_total_n_slices[epoch] = sum(epoch_to_ns_slices[epoch])
+    epoch_to_total_n_slices = {}
+    # [int]: list of epoch indices to sample from.
+    epoch_indices = []
+    # epoch_to_total_n_slices filtered using epoch_indices. It's an array for
+    # faster sampling.
+    sampling_epoch_weights = None
 
-    def extend_trajectory(t):
-      # TODO(pkozakowski) Refactor.
-      if len(t) == 1:
-        return t
-      else:
-        ts = t.timesteps[0]
-        extend_kwargs = {
-            'action': np.zeros_like(ts.action),
-            'dist_inputs': (
-                np.zeros_like(ts.dist_inputs)
-                if ts.dist_inputs is not None else None
-            ),
-            'reward': 0.0,
-            'done': True,
-            'mask': 0,
-            'new_observation': np.zeros_like(ts.observation),
-        }
-        t = t[:]  # Make a shallow copy of the timestep list.
-        for _ in range(margin):
-          t.extend(**extend_kwargs)
-        return t
-
-    while True:
+    def new_epoch(epoch_id):
+      """Updates the lists defined above to include the new epoch."""
       all_epochs = list(self._trajectories.keys())
       max_epoch = max(all_epochs) + 1
-      # Bind the epoch indices to a new name so they can be recalculated every
-      # epoch.
-      epoch_indices = epochs or all_epochs
-      epoch_indices = [
+
+      # Calculate the numbers of slices for the new epoch.
+      epoch_to_ns_slices[epoch_id] = np.array([
+          _n_slices(trajectory, max_slice_length, margin)
+          for trajectory in self._trajectories[epoch_id]
+      ])
+      epoch_to_total_n_slices[epoch_id] = np.sum(
+          epoch_to_ns_slices[epoch_id]
+      )
+
+      # Update the indices of epochs to sample from.
+      new_epoch_indices = epochs or all_epochs
+      new_epoch_indices = [
           # So -1 means "last".
-          ep % max_epoch for ep in epoch_indices
+          ep % max_epoch for ep in new_epoch_indices
       ]
       # Remove duplicates and consider only epochs where some trajectories
-      # were recorded.
-      epoch_indices = [epoch_id for epoch_id in list(set(epoch_indices))
-                       if self._trajectories[epoch_id]]
+      # were recorded and that we have processed in new_epoch.
+      new_epoch_indices = [
+          epoch_id for epoch_id in set(new_epoch_indices)
+          if self._trajectories[epoch_id] and epoch_id in epoch_to_ns_slices
+      ]
+      epoch_indices[:] = new_epoch_indices
+
+      nonlocal sampling_epoch_weights
+      sampling_epoch_weights = np.array(list(
+          epoch_to_total_n_slices[ep] for ep in epoch_indices
+      ))
+
+    while True:
+      # If we haven't collected any trajectories yet, yield an example
+      # trajectory. It's needed to determine the input/output shapes of
+      # networks.
+      if not self._trajectories:
+        yield self._random_slice(
+            self._example_trajectory, max_slice_length, margin
+        )
+        continue
+
+      # Catch up if we have a new epoch or we've restarted the experiment.
+      for epoch_id in self._trajectories.keys() - epoch_to_ns_slices.keys():  # pylint:disable=g-builtin-op
+        new_epoch(epoch_id)
 
       # Sample an epoch proportionally to number of slices in each epoch.
-      if len(epoch_indices) == 1:  # Skip this step if there's just 1 epoch.
-        epoch_id = epoch_indices[0]
-      else:
-        slices_per_epoch = [sum([n_slices(t) for t in self._trajectories[ep]])
-                            for ep in epoch_indices]
-        epoch_id = _sample_proportionally(epoch_indices, slices_per_epoch)
+      epoch_id = _sample_proportionally(epoch_indices, sampling_epoch_weights)
       epoch = self._trajectories[epoch_id]
 
       # Sample a trajectory proportionally to number of slices in each one.
       if sample_trajectories_uniformly:
-        slices_per_trajectory = [1] * len(epoch)
+        slices_per_trajectory = np.ones((len(epoch),))
       else:
-        slices_per_trajectory = [n_slices(t) for t in epoch]
+        slices_per_trajectory = epoch_to_ns_slices[epoch_id]
       trajectory = _sample_proportionally(epoch, slices_per_trajectory)
 
-      # Sample a slice from the trajectory.
-      slice_start = np.random.randint(n_slices(trajectory))
-      # Extend the trajectory with a given margin - this is to make sure that
-      # the networks always "see" the "done" states in the training data, even
-      # when a suffix is added to the trajectory slice for better estimation of
-      # returns.
-      extended_trajectory = extend_trajectory(trajectory)
-      slice_end = slice_start + (max_slice_length or len(extended_trajectory))
-      slice_end = min(slice_end, len(extended_trajectory) - end_offset)
-      yield extended_trajectory[slice_start:slice_end]
+      yield self._random_slice(trajectory, max_slice_length, margin)
 
   def trajectory_batch_stream(self, batch_size, epochs=None,
                               max_slice_length=None,
                               min_slice_length=None,
                               margin=0,
-                              include_final_state=False,
                               sample_trajectories_uniformly=False):
     """Return a stream of trajectory batches from the specified epochs.
 
@@ -686,8 +777,6 @@ class RLTask:
       min_slice_length: minimum length of the slices of trajectories to return
       margin: number of extra steps after "done" that should be included in
         slices, so that networks see the terminal states in the training data
-      include_final_state: whether to include slices with the final state of
-        the trajectory which may have no action and reward
       sample_trajectories_uniformly: whether to sample trajectories uniformly,
        or proportionally to the number of slices in each trajectory (default)
 
@@ -704,6 +793,8 @@ class RLTask:
       tensor_list = [t if t is not None else prototype for t in tensor_list]
 
       max_len = max([t.shape[0] for t in tensor_list])
+      if min_slice_length is not None:
+        max_len = max(max_len, min_slice_length)
       min_len = min([t.shape[0] for t in tensor_list])
       if max_len == min_len:  # No padding needed.
         return np.array(tensor_list)
@@ -713,35 +804,20 @@ class RLTask:
                        for t in tensor_list])
     cur_batch = []
     for t in self.trajectory_stream(
-        epochs, max_slice_length,
-        include_final_state, sample_trajectories_uniformly,
-        margin=margin
+        epochs, max_slice_length, sample_trajectories_uniformly, margin=margin
     ):
-      # TODO(pkozakowski): Instead sample the trajectories out of those with
-      # the minimum length.
-      if min_slice_length is not None and len(t) < min_slice_length:
-        continue
-
       cur_batch.append(t)
       if len(cur_batch) == batch_size:
-        # TODO(pkozakowski): Unpack based on name instead of position in the
-        # tuple (how?).
-        obs, act, dinp, rew, ret, done, mask = zip(*[
-            t.to_np(self._timestep_to_np) for t in cur_batch
-        ])
-        # Where act, rew and ret will usually have the following shape:
-        # [batch_size, trajectory_length-1], which we call [B, L-1].
-        # Observations are more complex and will usually be [B, L] + S where S
-        # is the shape of the observation space (self.observation_space.shape).
+        # bottleneck
+        # zip(*cur_batch) transposes (batch_size, fields)
+        # -> (fields, batch_size). Then we build TrajectoryNp from the fields.
+        # Fields are observations, actions, ...
+        batch_trajectory_np = TrajectoryNp(*zip(*cur_batch))
+        # Actions, rewards and returns in the trajectory slice have shape
+        # [batch_size, trajectory_length], which we denote as [B, L].
+        # Observations are more complex: [B, L] + S, where S is the shape of the
+        # observation space (self.observation_space.shape).
         # We stop the recursion at level 1, so we pass lists of arrays into
         # pad().
-        yield fastmath.nested_map(pad, TrajectoryNp(
-            observations=obs,
-            actions=act,
-            dist_inputs=dinp,
-            rewards=rew,
-            dones=done,
-            returns=ret,
-            mask=mask,
-        ), level=1)
+        yield fastmath.nested_map(pad, batch_trajectory_np, level=1)
         cur_batch = []
