@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -46,11 +46,13 @@ from trax.supervised import lr_schedules as lr
 class Agent:
   """Abstract class for RL agents, presenting the required API."""
 
-  def __init__(self, task: rl_task.RLTask,
+  def __init__(self,
+               task: rl_task.RLTask,
                n_trajectories_per_epoch=None,
                n_interactions_per_epoch=None,
                n_eval_episodes=0,
                eval_steps=None,
+               eval_temperatures=(0.0,),
                only_eval=False,
                output_dir=None,
                timestep_to_np=None):
@@ -68,15 +70,19 @@ class Agent:
         temperature 0 in each epoch -- used for evaluation only.
       eval_steps: an optional list of max_steps to use for evaluation
         (defaults to task.max_steps).
+      eval_temperatures: we always train with temperature 1 and evaluate with
+        temperature specified in the eval_temperatures list
+        (defaults to [0.0, 0.5])
       only_eval: If set to True, then trajectories are collected only for
         for evaluation purposes, but they are not recorded.
       output_dir: Path telling where to save outputs such as checkpoints.
       timestep_to_np: Timestep-to-numpy function to override in the task.
     """
-    assert bool(n_trajectories_per_epoch) != bool(n_interactions_per_epoch), (
-        'Exactly one of n_trajectories_per_epoch or n_interactions_per_epoch '
-        'should be specified.'
-    )
+    if n_trajectories_per_epoch is None == n_interactions_per_epoch is None:
+      raise ValueError(
+          'Exactly one of n_trajectories_per_epoch or '
+          'n_interactions_per_epoch should be specified.'
+      )
     self._epoch = 0
     self._task = task
     self._eval_steps = eval_steps or [task.max_steps]
@@ -88,7 +94,11 @@ class Agent:
     self._output_dir = output_dir
     self._avg_returns = []
     self._n_eval_episodes = n_eval_episodes
-    self._avg_returns_temperature0 = {step: [] for step in self._eval_steps}
+    self._eval_temperatures = eval_temperatures
+    self._avg_returns_temperatures = {
+        eval_t: {step: [] for step in self._eval_steps
+                } for eval_t in eval_temperatures
+    }
     if self._output_dir is not None:
       self.init_from_file()
 
@@ -124,9 +134,10 @@ class Agent:
     task_path = os.path.join(self._output_dir, task_file_name)
     self._task.save_to_file(task_path)
     file_path = os.path.join(self._output_dir, file_name)
-    dictionary = {'epoch': self._epoch,
-                  'avg_returns': self._avg_returns,
-                  'avg_returns_temperature0': self._avg_returns_temperature0}
+    dictionary = {'epoch': self._epoch, 'avg_returns': self._avg_returns}
+    for eval_t in self._eval_temperatures:
+      dictionary['avg_returns_temperature_{}'.format(
+          eval_t)] = self._avg_returns_temperatures[eval_t]
     with tf.io.gfile.GFile(file_path, 'wb') as f:
       pickle.dump(dictionary, f)
 
@@ -144,7 +155,9 @@ class Agent:
       dictionary = pickle.load(f)
     self._epoch = dictionary['epoch']
     self._avg_returns = dictionary['avg_returns']
-    self._avg_returns_temperature0 = dictionary['avg_returns_temperature0']
+    for eval_t in self._eval_temperatures:
+      self._avg_returns_temperatures[eval_t] = dictionary[
+          'avg_returns_temperature_{}'.format(eval_t)]
 
   def _collect_trajectories(self):
     return self.task.collect_trajectories(
@@ -226,25 +239,29 @@ class Agent:
             'Average return in epoch %d was %.2f.' % (self._epoch, avg_return))
         if self._n_eval_episodes > 0:
           for steps in self._eval_steps:
-            avg_return_temperature0 = self.task.collect_trajectories(
-                lambda x: self.policy(x, temperature=0.0),
-                n_trajectories=self._n_eval_episodes,
-                max_steps=steps, only_eval=True)
-            self._avg_returns_temperature0[steps].append(
-                avg_return_temperature0
-            )
-            supervised.trainer_lib.log(
-                'Avg return with temperature 0 at %d steps in epoch %d was %.2f.'
-                % (steps, self._epoch, avg_return_temperature0))
+            for eval_t in self._eval_temperatures:
+              avg_return_temperature = self.task.collect_trajectories(
+                  functools.partial(self.policy, temperature=eval_t),
+                  n_trajectories=self._n_eval_episodes,
+                  max_steps=steps,
+                  only_eval=True)
+              supervised.trainer_lib.log(
+                  'Eval return in epoch %d with temperature %.2f was %.2f.'
+                  % (self._epoch, eval_t, avg_return_temperature))
+              self._avg_returns_temperatures[eval_t][steps].append(
+                  avg_return_temperature)
+
         if sw is not None:
           sw.scalar('timing/collect', time.time() - cur_time,
                     step=self._epoch)
           sw.scalar('rl/avg_return', avg_return, step=self._epoch)
           if self._n_eval_episodes > 0:
             for steps in self._eval_steps:
-              sw.scalar('rl/avg_return_temperature0_steps%d' % steps,
-                        self._avg_returns_temperature0[steps][-1],
-                        step=self._epoch)
+              for eval_t in self._eval_temperatures:
+                sw.scalar(
+                    'rl/avg_return_temperature%.2f_steps%d' % (eval_t, steps),
+                    self._avg_returns_temperatures[eval_t][steps][-1],
+                    step=self._epoch)
           sw.scalar('rl/n_interactions', self.task.n_interactions(),
                     step=self._epoch)
           sw.scalar('rl/n_interactions_per_second',
@@ -344,7 +361,6 @@ class PolicyAgent(Agent):
         policy_model(mode='eval'), n_devices=1)  # Not collecting stats
     self._policy_eval_model.init(shapes.signature(policy_batch))
     if self._task._initial_trajectories == 0:
-      self._task.remove_epoch(0)
       self._collect_trajectories()
 
   @property
@@ -495,6 +511,7 @@ class PolicyGradient(Agent):
         checkpoint_at=checkpoint_at,
     )
     self._collect_model = model_fn(mode='collect')
+    self._collect_model.init(shapes.signature(train_task.sample_batch))
 
     # Validate the restored checkpoints. The number of network training steps
     # (self.loop.step) should be equal to the number of epochs (self._epoch),
@@ -525,6 +542,8 @@ class PolicyGradient(Agent):
 
   def train_epoch(self):
     """Trains RL for one epoch."""
+    # Copy policy state accumulated during data collection to the trainer.
+    self._loop.update_weights_and_state(state=self._collect_model.state)
     # Perform one gradient step per training epoch to ensure we stay on policy.
     self._loop.run(n_steps=1)
 
@@ -537,7 +556,12 @@ class PolicyGradient(Agent):
 
 
 def network_policy(
-    collect_model, policy_distribution, loop, trajectory_np, temperature=1.0
+    collect_model,
+    policy_distribution,
+    loop,
+    trajectory_np,
+    head_index=0,
+    temperature=1.0,
 ):
   """Policy function powered by a neural network.
 
@@ -548,6 +572,7 @@ def network_policy(
     policy_distribution: an instance of trax.rl.distributions.Distribution
     loop: trax.supervised.training.Loop used to train the policy network
     trajectory_np: an instance of trax.rl.task.TrajectoryNp
+    head_index: index of the policy head a multihead model.
     temperature: temperature used to sample from the policy (default=1.0)
 
   Returns:
@@ -567,10 +592,13 @@ def network_policy(
   # Copying weights from loop.model should work, because the raw model's
   # weights should be updated automatically during training, but it doesn't.
   # TODO(pkozakowski): Debug.
-  acc = loop._trainer_per_task[0].accelerated_loss_layer  # pylint: disable=protected-access
+  acc = loop._trainer_per_task[0].accelerated_model_with_loss  # pylint: disable=protected-access
   model.weights = acc._unreplicate(acc.weights[0])  # pylint: disable=protected-access
   # Add batch dimension to trajectory_np and run the model.
   pred = model(trajectory_np.observations[None, ...])
+  if isinstance(pred, (tuple, list)):
+    # For multihead models, extract the policy head output.
+    pred = pred[head_index]
   assert pred.shape == (
       1, trajectory_np.observations.shape[0], policy_distribution.n_inputs
   )
@@ -683,7 +711,8 @@ class ValueAgent(Agent):
     self._value_eval_model = value_model(mode='eval')
     self._value_eval_model.init(self._value_model_signature)
     self._value_eval_jit = tl.jit_forward(
-        self._value_eval_model.pure_fn, fastmath.device_count(), do_mean=False)
+        self._value_eval_model.pure_fn, fastmath.local_device_count(),
+        do_mean=False)
 
     # Inputs to the value model are produced by self._values_batches_stream.
     self._inputs = data.inputs.Inputs(
@@ -708,7 +737,6 @@ class ValueAgent(Agent):
         value_model(mode='collect'), n_devices=1)
     self._eval_model.init(shapes.signature(value_batch))
     if self._task._initial_trajectories == 0:
-      self._task.remove_epoch(0)
       self._collect_trajectories()
 
   @property
@@ -935,7 +963,7 @@ class DQN(ValueAgent):
 
   def _run_value_model(self, obs, use_eval_model=True):
     """Runs value model."""
-    n_devices = fastmath.device_count()
+    n_devices = fastmath.local_device_count()
     if use_eval_model:
       weights = tl.for_n_devices(self._value_eval_model.weights, n_devices)
       state = tl.for_n_devices(self._value_eval_model.state, n_devices)
