@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -47,7 +47,7 @@ class DecodingTest(test.TestCase):
         model, batch_size=1, eos_id=-1, max_length=10)
     self.assertEqual(s1.shape[0], 1)
     self.assertEqual(s1.shape[1], 10)
-    batch_per_device = 2 // fastmath.device_count()
+    batch_per_device = 2 // fastmath.local_device_count()
     model.init(shapes.ShapeDtype((batch_per_device, 1), dtype=np.int32))
     s2 = decoding.autoregressive_sample(
         model, batch_size=2, max_length=10)
@@ -60,6 +60,28 @@ class DecodingTest(test.TestCase):
     self.assertEqual(s3.shape[0], 1)
     self.assertEqual(s3.shape[1], 10)
 
+  def test_autoregressive_sample_transformerlm_tfnp(self):
+    with fastmath.use_backend(fastmath.Backend.TFNP):
+      model = models.TransformerLM(10, d_model=32, d_ff=64, n_layers=1,
+                                   n_heads=2, mode='predict')
+      model.init(shapes.ShapeDtype((1, 1), dtype=np.int32))
+      s1 = decoding.autoregressive_sample(
+          model, batch_size=1, eos_id=-1, max_length=10)
+      self.assertEqual(s1.shape[0], 1)
+      self.assertEqual(s1.shape[1], 10)
+      batch_per_device = 2 // fastmath.local_device_count()
+      model.init(shapes.ShapeDtype((batch_per_device, 1), dtype=np.int32))
+      s2 = decoding.autoregressive_sample(
+          model, batch_size=2, max_length=10)
+      self.assertEqual(s2.shape[0], 2)
+      self.assertLess(s2.shape[1], 11)
+      model.init(shapes.ShapeDtype((1, 1), dtype=np.int32))
+      prefix = np.array([[1, 2, 3]])
+      s3 = decoding.autoregressive_sample(model, prefix, eos_id=-1,
+                                          max_length=10, batch_size=1)
+      self.assertEqual(s3.shape[0], 1)
+      self.assertEqual(s3.shape[1], 10)
+
   def _lsh_self_attention_fn(self):
     return functools.partial(
         tl.LSHSelfAttention,
@@ -74,6 +96,24 @@ class DecodingTest(test.TestCase):
         predict_mem_len=1024,
     )
 
+  def _pure_lsh_self_attention_fn(self, n_chunks_after=0):
+    return functools.partial(
+        tl.PureLSHSelfAttentionWrapper,
+        attention_dropout=0.0,
+        chunk_len=16,
+        n_buckets=[32, 32],
+        n_chunks_after=n_chunks_after,
+        n_chunks_before=1,
+        n_hashes=2,
+        n_parallel_heads=1,
+        max_length_for_buckets=1024,
+        predict_drop_len=128,
+        predict_mem_len=1024,
+        num_weights=2,
+        bias=False,
+        pure_lsh_implementation=tl.PureLSHSelfAttention,
+    )
+
   def _timebin_self_attention_fn(self, use_reference_code=False, chunk_len=64):
     return functools.partial(
         tl.SelfAttention,
@@ -101,8 +141,8 @@ class DecodingTest(test.TestCase):
                               max_len=65536,
                               attention_type=[timebin_self_attention,
                                               lsh_self_attention],
-                              axial_pos_shape=(256, 256),
-                              d_axial_pos_embs=(128, 128),
+                              pos_axial_shape=(256, 256),
+                              pos_d_axial_embs=(128, 128),
                               ff_activation=tl.Relu,
                               ff_use_sru=0,
                               mode='predict',
@@ -137,20 +177,20 @@ class DecodingTest(test.TestCase):
                                        max_length=6, temperature=0.0)
     self.assertEqual(str(s[0]), '[3 7 5 3 2 4]')
 
-  def test_autoregressive_sample_reformerlm_quality(self):
-    timebin_self_attention = self._timebin_self_attention_fn()
-    pred_model = models.ReformerLM(
+  def test_autoregressive_sample_transformerlm_quality_beam(self):
+    pred_model = models.TransformerLM(
         d_model=64, d_ff=128, dropout=0.05, max_len=256, n_heads=2,
-        attention_type=timebin_self_attention,
         n_layers=2, vocab_size=13, mode='predict')
     shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
-    model_path = os.path.join(_TESTDATA, 'reformerlm_copy.pkl.gz')
+    model_path = os.path.join(_TESTDATA, 'transformerlm_copy.pkl.gz')
     pred_model.init_from_file(model_path, weights_only=True,
                               input_signature=(shape11, shape11))
     inputs = np.array([[0, 3, 7, 5, 3, 2, 4, 0]], dtype=np.int32)
-    s = decoding.autoregressive_sample(pred_model, inputs,
-                                       max_length=6, temperature=0.0)
-    self.assertEqual(str(s[0]), '[3 7 5 3 2 4]')
+    s = decoding.beam_search(pred_model, inputs, n_beams=3, max_length=6)
+    self.assertEqual(len(s), 3)  # 3 beams
+    self.assertEqual(str(s[0][0][0]), '[3 7 5 3 2 4]')
+    self.assertEqual(str(s[1][0][0]), '[3 7 5 3 2 2]')  # different from above
+    self.assertEqual(str(s[2][0][0]), '[3 7 5 3 3 2]')  # different from above
 
   def test_autoregressive_sample_transformer_quality(self):
     pred_model = models.Transformer(
@@ -184,7 +224,7 @@ class DecodingTest(test.TestCase):
         encoder_attention_type=self._lsh_self_attention_fn(),
         encoder_decoder_attention_type=self._lsh_self_attention_fn(),
         input_vocab_size=256,
-        axial_pos_shape=None,
+        pos_axial_shape=None,
     )
 
     shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
@@ -205,15 +245,16 @@ class DecodingTest(test.TestCase):
 
   def test_autoregressive_sample_reformer2_lsh_attn_quality(self):
     gin.add_config_file_search_path(_CONFIG_DIR)
-    # 32 is the max length we trained the checkpoint for.
+    max_len = 32  # 32 is the max length we trained the checkpoint for.
     test_lengths = [8, 16, 32]
     vocab_size = 13
+    # The checkpoint is correct on ~90% sequences, set random seed to deflake.
     np.random.seed(0)
-    for max_len in test_lengths:
+    for test_len in test_lengths:
       gin.clear_config()
       gin.parse_config_file('reformer2_copy.gin')
       gin.bind_parameter('LSHSelfAttention.predict_mem_len', 2 * max_len)
-      gin.bind_parameter('LSHSelfAttention.predict_drop_len', max_len // 4)
+      gin.bind_parameter('LSHSelfAttention.predict_drop_len', 2 * max_len)
 
       pred_model = models.Reformer2(mode='predict')
 
@@ -225,12 +266,12 @@ class DecodingTest(test.TestCase):
                                 input_signature=(shape1l, shape11))
       initial_state = pred_model.state
 
-      for _ in range(3):
-        # pick a length in [1, max_len]
-        inp_len = np.random.randint(low=1, high=max_len + 1)
-        inputs = np.random.randint(low=1, high=vocab_size-1, size=(1, inp_len))
-        inputs = np.pad(inputs, [(0, 0), (0, max_len - inp_len)],
-                        mode='constant', constant_values=0)
+      for _ in range(2):  # Set low to make the test run reasonably fast.
+        # Pick a length in [1, test_len] at random.
+        inp_len = np.random.randint(low=1, high=test_len + 1)
+        inputs = np.random.randint(low=1, high=vocab_size-1, size=(1, max_len))
+        # TODO(jaszczur): properly fix padding in Reformer2 predict mode,
+        # and add a test here.
         s = decoding.autoregressive_sample(
             pred_model, inputs=inputs, eos_id=-1, max_length=inp_len,
             temperature=0.0)
@@ -238,52 +279,6 @@ class DecodingTest(test.TestCase):
         pred_model.state = initial_state
     gin.clear_config()  # Make sure to not affect other tests.
 
-  def test_autoregressive_sample_reformer2_copy_self_attn_quality(self):
-    max_len = 32
-
-    def _self_attention_fn():
-      return functools.partial(
-          tl.SelfAttention,
-          predict_drop_len=2 * max_len,
-          predict_mem_len=2 * max_len,
-      )
-
-    pred_model = models.Reformer2(
-        mode='predict',
-        d_model=256,
-        d_ff=512,
-        dropout=0.05,
-        max_len=max_len,
-        n_heads=4,
-        n_encoder_layers=3,
-        n_decoder_layers=3,
-        ff_use_sru=1,
-        d_attention_key=64,
-        d_attention_value=64,
-        encoder_attention_type=_self_attention_fn(),
-        encoder_decoder_attention_type=_self_attention_fn(),
-        input_vocab_size=13,
-        axial_pos_shape=None,
-    )
-
-    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
-    shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)
-
-    model_path = os.path.join(_TESTDATA,
-                              'reformer2_copy_self_attn.pkl.gz')
-    pred_model.init_from_file(model_path, weights_only=True,
-                              input_signature=(shape1l, shape11))
-
-    inputs = np.array([[11, 5, 1, 2, 3, 4]], dtype=np.int32)
-    inp_len = inputs.shape[1]
-    inputs = np.pad(inputs, [(0, 0), (0, max_len - inp_len)],
-                    mode='constant', constant_values=0)
-    s = decoding.autoregressive_sample(
-        pred_model, inputs=inputs, eos_id=-1, max_length=inp_len,
-        temperature=0.0)
-
-    np.testing.assert_equal(s[0], inputs[0, :inp_len])
-
   def test_autoregressive_sample_reformerlm_lsh(self):
     max_len = 32
 
@@ -306,8 +301,8 @@ class DecodingTest(test.TestCase):
                                          predict_mem_len=max_len,
                                          max_length_for_buckets=1024),
         vocab_size=13,
-        axial_pos_shape='fixed-base',
-        d_axial_pos_embs=None,
+        pos_type='fixed-base',
+        pos_d_axial_embs=None,
     )
 
     shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
@@ -345,8 +340,8 @@ class DecodingTest(test.TestCase):
                                          predict_mem_len=max_len,
                                          max_length_for_buckets=1024),
         vocab_size=13,
-        axial_pos_shape='fixed-base',
-        d_axial_pos_embs=None,
+        pos_type='fixed-base',
+        pos_d_axial_embs=None,
     )
 
     shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
@@ -366,6 +361,83 @@ class DecodingTest(test.TestCase):
     np.testing.assert_equal(s[0], inputs[0, 1:inp_len-1])
     # pylint: enable=unreachable
 
+  def test_autoregressive_sample_reformer2_pure_lsh(self):
+    max_len = 128
+
+    pred_model = models.Reformer2(
+        mode='predict',
+        d_model=256,
+        d_ff=512,
+        dropout=0.05,
+        max_len=max_len,
+        n_heads=4,
+        n_encoder_layers=1,
+        n_decoder_layers=1,
+        ff_use_sru=1,
+        d_attention_key=64,
+        d_attention_value=64,
+        encoder_attention_type=self._pure_lsh_self_attention_fn(
+            n_chunks_after=1),
+        encoder_decoder_attention_type=self._pure_lsh_self_attention_fn(),
+        input_vocab_size=256,
+        pos_axial_shape=None,
+    )
+
+    shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
+    shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)
+    pred_model.init(input_signature=(shape1l, shape11))
+
+    # 0w0w
+    inputs = np.array(
+        [[0, 3, 7, 5, 3, 2, 4, 1, 8, 0, 3, 7, 5, 3, 2, 4, 1, 8]],
+        dtype=np.int32)
+    inputs = np.pad(inputs, [(0, 0), (0, max_len - inputs.shape[1])],
+                    mode='constant', constant_values=0)
+    s = decoding.autoregressive_sample(
+        pred_model, inputs=inputs, eos_id=-1, max_length=10, temperature=0.0)
+
+    self.assertEqual(s.shape[0], 1)
+    self.assertEqual(s.shape[1], 10)
+
+  def test_autoregressive_sample_reformer2_pure_lsh_attn_quality(self):
+    gin.add_config_file_search_path(_CONFIG_DIR)
+    max_len = 32  # 32 is the max length we trained the checkpoint for.
+    test_lengths = [8, 16, 32]
+    vocab_size = 13
+    # The checkpoint is correct on ~90% sequences, set random seed to deflake.
+    np.random.seed(0)
+    for test_len in test_lengths:
+      gin.clear_config()
+      gin.parse_config_file('reformer2_purelsh_copy.gin')
+      gin.bind_parameter('PureLSHSelfAttention.predict_mem_len', 2 * max_len)
+      gin.bind_parameter('PureLSHSelfAttention.predict_drop_len', 2 * max_len)
+      gin.bind_parameter('PureLSHSelfAttentionWrapper.bias', False)
+      gin.bind_parameter('PureLSHSelfAttentionWrapper.num_weights', 2)
+
+      pred_model = models.Reformer2(mode='predict')
+
+      shape11 = shapes.ShapeDtype((1, 1), dtype=np.int32)
+      shape1l = shapes.ShapeDtype((1, max_len), dtype=np.int32)
+
+      model_path = os.path.join(_TESTDATA, 'reformer2_purelsh_copy.pkl.gz')
+      pred_model.init_from_file(model_path, weights_only=True,
+                                input_signature=(shape1l, shape11))
+      initial_state = pred_model.state
+
+      for _ in range(2):  # Set low to make the test run reasonably fast.
+        # Pick a length in [1, test_len] at random.
+        inp_len = np.random.randint(low=1, high=test_len + 1)
+        inputs = np.random.randint(low=1, high=vocab_size-1, size=(1, max_len))
+        # TODO(jaszczur): properly fix padding in Reformer2 predict mode,
+        # and add a test here.
+        s = decoding.autoregressive_sample(
+            pred_model, inputs=inputs, eos_id=-1, max_length=inp_len,
+            temperature=0.0)
+
+        np.testing.assert_equal(s[0], inputs[0, :inp_len])
+        pred_model.state = initial_state
+    gin.clear_config()  # Make sure to not affect other tests.
+
 
 if __name__ == '__main__':
   config.config_with_absl()
