@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -17,19 +17,18 @@
 
 import functools
 import os
+
 from absl.testing import absltest
 from absl.testing import parameterized
-
 from jax import test_util  # pylint: disable=unused-import
 from jax.config import config
 from jax.lib import xla_bridge
-
 import tensorflow.compat.v2 as tf
-
 from trax import fastmath
 from trax import layers as tl
 from trax import models
 from trax import optimizers as trax_opt
+from trax import shapes as trax_shapes
 from trax import test_utils
 from trax.data import inputs as inputs_lib
 from trax.fastmath import numpy as jnp
@@ -90,7 +89,8 @@ def _test_inputs_lm(vocab_size, seq_len, per_device_batch_size=2):
 
 
 
-BACKENDS = [fastmath.Backend.JAX, fastmath.Backend.TFNP]
+BACKENDS = [fastmath.Backend.JAX]
+BACKENDS_AND_CONFIGS = [(fastmath.Backend.JAX, [('Simple', None)])]
 
 
 def short_name(b):
@@ -106,6 +106,44 @@ def opt_name(opt):
   return opt.__name__
 
 
+def _pure_lsh_self_attention_fn(n_chunks_after=0):
+  return functools.partial(
+      tl.PureLSHSelfAttentionWrapper,
+      attention_dropout=0.1,
+      chunk_len=16,
+      n_buckets=[32, 32],
+      n_chunks_after=n_chunks_after,
+      n_chunks_before=1,
+      n_hashes=2,
+      n_parallel_heads=1,
+      max_length_for_buckets=1024,
+      predict_drop_len=128,
+      predict_mem_len=1024,
+      num_weights=2,
+      bias=False,
+      pure_lsh_implementation=tl.PureLSHSelfAttention,
+  )
+
+
+def _mixed_lsh_self_attention_fn(n_chunks_after=0):
+  return functools.partial(
+      tl.PureLSHSelfAttentionWrapper,
+      attention_dropout=0.1,
+      chunk_len=16,
+      n_buckets=[32, 32],
+      n_chunks_after=n_chunks_after,
+      n_chunks_before=1,
+      n_hashes=2,
+      n_parallel_heads=1,
+      max_length_for_buckets=1024,
+      predict_drop_len=128,
+      predict_mem_len=1024,
+      num_weights=2,
+      bias=False,
+      pure_lsh_implementation=tl.MixedLSHSelfAttention,
+  )
+
+
 class TraxTest(parameterized.TestCase):
 
   def __init__(self, methodName='runTest'):  # pylint: disable=invalid-name
@@ -137,8 +175,9 @@ class TraxTest(parameterized.TestCase):
         # Adds Dropout and BatchNorm to test state handling.
         def model_fn(mode='train'):
           return tl.Serial(
-              tl.Dropout(mode=mode, rate=0.1), tl.BatchNorm(mode=mode),
-              models.MLP(d_hidden=16, n_output_classes=n_classes, mode=mode))
+              tl.Dropout(mode=mode, rate=0.1),
+              tl.BatchNorm(mode=mode),
+              models.MLP(layer_widths=(16, 16, n_classes), mode=mode))
         inputs = _test_inputs(n_classes)
         n_in = 1
       elif model_name == 'Resnet50':
@@ -191,11 +230,7 @@ class TraxTest(parameterized.TestCase):
   @parameterized.named_parameters(
       ('_%s_%s_%s' % (short_name(backend), model_name, opt_name(opt)),  # pylint: disable=g-complex-comprehension
        backend, model_name, opt)
-      for backend, configs in [
-          (fastmath.Backend.JAX, [('Simple', None)]),
-          (fastmath.Backend.TFNP, [('Simple', None),
-                                   ('Resnet50', trax_opt.Momentum),
-                                   ('Transformer', trax_opt.Adam)])]
+      for backend, configs in BACKENDS_AND_CONFIGS
       for model_name, opt in configs)
   def test_train_eval_predict(self, backend, model_name, opt):
     self._test_train_eval_predict(backend, model_name, opt)
@@ -211,8 +246,8 @@ class TraxTest(parameterized.TestCase):
       n_classes = 4
       steps = 2
       eval_steps = 2
-      model_fn = functools.partial(
-          models.MLP, d_hidden=16, n_output_classes=n_classes)
+      model_fn = functools.partial(models.MLP,
+                                   layer_widths=(16, 16, n_classes))
       inputs = _test_inputs(n_classes)
 
       # Train and evaluate
@@ -234,19 +269,222 @@ class TraxTest(parameterized.TestCase):
           eval_steps=eval_steps,
           eval_frequency=1)
 
-      # Assert total train steps - with loop we don't resume, but train for as
-      # many steps as given, so: steps + 2*steps = 3*steps.
-      self.assertEqual(loop.step, 3 * steps)
+      # Assert total train steps
+      self.assertEqual(loop.step, 2 * steps)
 
   @parameterized.parameters(BACKENDS)
+  def test_train_permanent_checkpoints(self, backend):
+    with fastmath.use_backend(backend):
+      # Prepare model and inputs
+      n_classes = 4
+      steps = 5
+      eval_steps = 2
+      model_fn = functools.partial(models.MLP,
+                                   layer_widths=(16, 16, n_classes))
+      inputs = _test_inputs(n_classes)
+
+      # Train and evaluate
+      output_dir = self.create_tempdir().full_path
+
+      # Steps 1 -> 5
+      loop = trainer_lib.train(
+          output_dir,
+          model=model_fn,
+          inputs=inputs,
+          steps=steps,
+          eval_steps=eval_steps,
+          eval_frequency=1,
+          permanent_checkpoint_frequency=2)
+
+      # Steps 6 -> 10
+      loop = trainer_lib.train(
+          output_dir,
+          model=model_fn,
+          inputs=inputs,
+          steps=(2 * steps),
+          eval_steps=eval_steps,
+          eval_frequency=1,
+          permanent_checkpoints_at=[7, 8, 10])
+
+      path = os.path.join(output_dir, 'model.pkl.gz')
+      self.assertTrue(tf.io.gfile.exists(path))
+
+      for step in range(11):
+        filename = 'model_{}.pkl.gz'.format(step)
+        path = os.path.join(output_dir, filename)
+        if step in [1, 2, 4, 7, 8, 10]:
+          self.assertTrue(tf.io.gfile.exists(path),
+                          msg='No model for step: {} in dir {}.'.format(
+                              step, tf.io.gfile.listdir(output_dir)))
+        else:
+          self.assertFalse(tf.io.gfile.exists(path),
+                           msg='Model for step: {} in dir {}.'.format(
+                               step, tf.io.gfile.listdir(output_dir)))
+
+      # Assert total train steps
+      self.assertEqual(loop.step, 10)
+
+  @parameterized.parameters(BACKENDS)
+  def test_train_restart_with_same_steps(self, backend):
+    with fastmath.use_backend(backend):
+      # Prepare model and inputs
+      n_classes = 4
+      steps = 2
+      eval_steps = 2
+      model_fn = functools.partial(models.MLP,
+                                   layer_widths=(16, 16, n_classes))
+      inputs = _test_inputs(n_classes)
+
+      # Train and evaluate
+      output_dir = self.create_tempdir().full_path
+      trainer_lib.train(
+          output_dir,
+          model=model_fn,
+          inputs=inputs,
+          steps=steps,
+          eval_steps=eval_steps,
+          eval_frequency=1)
+
+      # Restart training
+      loop = trainer_lib.train(
+          output_dir,
+          model=model_fn,
+          inputs=inputs,
+          steps=steps,
+          eval_steps=eval_steps,
+          eval_frequency=1)
+
+      # Assert total train steps
+      self.assertEqual(loop.step, steps)
+
+  def test_train_with_pure_lsh_attention(self, backend=fastmath.Backend.JAX):
+    with fastmath.use_backend(backend):
+      # Prepare model and inputs
+      def model(mode='train'):
+        return models.Reformer2(
+            mode=mode,
+            d_model=16,
+            d_ff=16,
+            n_heads=2,
+            dropout=0.05,
+            n_decoder_layers=1,
+            n_encoder_layers=1,
+            input_vocab_size=256,
+            encoder_attention_type=_pure_lsh_self_attention_fn(),
+            encoder_decoder_attention_type=_pure_lsh_self_attention_fn(),
+        )
+
+      max_len = 128
+      inputs = _test_inputs_lm(vocab_size=256, seq_len=max_len)
+
+      steps = 1
+      eval_steps = 1
+
+      # Train and evaluate
+      output_dir = self.create_tempdir().full_path
+      trainer_lib.train(
+          output_dir,
+          model=model,
+          inputs=inputs,
+          steps=steps,
+          eval_steps=eval_steps,
+          eval_frequency=1)
+
+      # Read checkpoint
+      model_file = os.path.join(output_dir, 'model.pkl.gz')
+
+      shape11 = trax_shapes.ShapeDtype((1, 1), dtype=jnp.int32)
+      shape1l = trax_shapes.ShapeDtype((1, max_len), dtype=jnp.int32)
+
+      model_predict = model(mode='predict')
+      model_predict.init_from_file(
+          model_file, weights_only=True, input_signature=(shape1l, shape11))
+
+  def test_train_with_mixed_lsh_attention(self, backend=fastmath.Backend.JAX):
+    with fastmath.use_backend(backend):
+      # Prepare model and inputs
+
+      def model(mode='train'):
+        return models.Reformer2(
+            mode=mode,
+            d_model=16,
+            d_ff=16,
+            n_heads=2,
+            dropout=0.05,
+            n_decoder_layers=1,
+            n_encoder_layers=1,
+            input_vocab_size=256,
+            encoder_attention_type=_mixed_lsh_self_attention_fn(),
+            encoder_decoder_attention_type=_mixed_lsh_self_attention_fn(),
+        )
+
+      max_len = 128
+      inputs = _test_inputs_lm(vocab_size=256, seq_len=max_len)
+
+      steps = 1
+      eval_steps = 1
+
+      # Train and evaluate
+      output_dir = self.create_tempdir().full_path
+      trainer_lib.train(
+          output_dir,
+          model=model,
+          inputs=inputs,
+          steps=steps,
+          eval_steps=eval_steps,
+          eval_frequency=1)
+
+      # Read checkpoint
+      model_file = os.path.join(output_dir, 'model.pkl.gz')
+
+      shape11 = trax_shapes.ShapeDtype((1, 1), dtype=jnp.int32)
+      shape1l = trax_shapes.ShapeDtype((1, max_len), dtype=jnp.int32)
+
+      model_predict = model(mode='predict')
+      model_predict.init_from_file(model_file, weights_only=True,
+                                   input_signature=(shape1l, shape11))
+
+  @parameterized.parameters(BACKENDS)
+  def test_train_fills_in_missing_eval_metrics(self, backend):
+    with fastmath.use_backend(backend):
+      # Prepare model and inputs
+      n_classes = 4
+      steps = 2
+      eval_steps = 2
+      model_fn = functools.partial(models.MLP, layer_widths=(16, 16, n_classes))
+      inputs = _test_inputs(n_classes)
+      additional_eval_stream = trainer_lib.NamedStream(
+          # deliberately duplicating eval data
+          stream=inputs.eval_stream(1),
+          name='additional_eval_task')
+
+      # Train and evaluate
+      output_dir = self.create_tempdir().full_path
+      loop = trainer_lib.train(
+          output_dir,
+          model=model_fn,
+          inputs=inputs,
+          steps=steps,
+          eval_steps=eval_steps,
+          eval_frequency=1,
+          additional_eval_streams=[additional_eval_stream])
+
+      self.assertLen(loop.eval_tasks, 2)
+      eval_task_1, eval_task_2 = loop.eval_tasks
+      self.assertCountEqual(eval_task_1.metrics, eval_task_2.metrics)
+      self.assertCountEqual(eval_task_1.metric_names, eval_task_2.metric_names)
+
+  @parameterized.named_parameters(
+      ('_%s' % short_name(backend), backend)
+      for backend in BACKENDS)
   def test_train_with_weights(self, backend):
     with fastmath.use_backend(backend):
       # Prepare model and inputs
       n_classes = 4
       steps = 2
       eval_steps = 2
-      model_fn = functools.partial(
-          models.MLP, d_hidden=16, n_output_classes=n_classes)
+      model_fn = functools.partial(models.MLP,
+                                   layer_widths=(16, 16, n_classes))
       inputs = _test_inputs(n_classes, with_weights=True)
 
       # Train and evaluate
@@ -265,13 +503,13 @@ class TraxTest(parameterized.TestCase):
   def test_reset_twice(self, backend):
     with fastmath.use_backend(backend):
       n_classes = 4
-      model_fn = functools.partial(
-          models.MLP, d_hidden=16, n_output_classes=n_classes)
+      model_fn = functools.partial(models.MLP,
+                                   layer_widths=(16, 16, n_classes))
       inputs = _test_inputs(n_classes)
 
       trainer = trainer_lib.Trainer(
           model=model_fn,
-          loss_fn=tl.CrossEntropyLoss(),
+          loss_fn=tl.WeightedCategoryCrossEntropy(),
           optimizer=trax_opt.SM3,
           lr_schedule=lr.multifactor(),
           inputs=inputs,
@@ -287,41 +525,10 @@ class TraxTest(parameterized.TestCase):
   def test_tf_xla_forced_compile(self):
     # TODO(wangpeng): re-enable this test
     self.skipTest('Needs --config=cuda to pass this test')
-    old_flag = fastmath.tf_math.tf_xla_forced_compile_enabled()
-    fastmath.tf_math.set_tf_xla_forced_compile(True)
+    old_flag = fastmath.tf.tf_xla_forced_compile_enabled()
+    fastmath.tf.set_tf_xla_forced_compile(True)
     self._test_train_eval_predict('tf')
-    fastmath.tf_math.set_tf_xla_forced_compile(old_flag)
-
-  def test_no_int32_or_uint32_returned(self):
-    """Tests that Trainer._jit_update_fn doesn't return int32 or uint32.
-
-    TF pins int32/uint32 tensors to CPU, which will cause XLA-forced-compiled
-    computation to copy int32/uint32 outputs to CPU. This test makes sure that
-    won't happen.
-    """
-    with fastmath.use_backend(fastmath.Backend.TFNP):
-      n_classes = 1001
-      model_fn = functools.partial(models.Resnet50,
-                                   n_output_classes=n_classes)
-      inputs = _test_inputs(n_classes, input_shape=(224, 224, 3))
-      trainer = trainer_lib.Trainer(
-          model=model_fn,
-          loss_fn=tl.CrossEntropyLoss(),
-          optimizer=trax_opt.SM3,
-          lr_schedule=lr.multifactor(),
-          inputs=inputs,
-      )
-      output_dir = self.create_tempdir().full_path
-      trainer.reset(output_dir)
-      trainer.train_epoch(1, 0)
-      # Those are the things returned by Trainer._jit_update_fn
-      arrays = (trainer._opt_state.weights, trainer._opt_state.slots,
-                trainer._model_state, trainer._rngs)
-      arrays = tf.nest.flatten(arrays)
-      for x in arrays:
-        if isinstance(x, jnp.ndarray) and (x.dtype == jnp.int32 or
-                                           x.dtype == jnp.uint32):
-          raise ValueError('Found an array of int32 or uint32: %s' % x)
+    fastmath.tf.set_tf_xla_forced_compile(old_flag)
 
 
 
