@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,11 +14,6 @@
 # limitations under the License.
 
 """Extensions such as `jit`, `grad`, `logsumexp`, etc."""
-
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import bisect
 import contextlib
 import copy
@@ -57,40 +52,39 @@ def most_precise_int_dtype(x):
 
 def _canonicalize_jit_arg(x):
   if isinstance(x, tf_np.ndarray):
-    return x.data
-  else:
-    try:
-      # We need to convert `int` to the most precise dtype, otherwise the dtype
-      # of the result may be different from numpy's. For example, when a binary
-      # op takes in a Python integer 5 and an array of uint32, numpy will pick
-      # uint32 as 5's dtype, while tf.convert_to_tensor will choose int32 which
-      # will cause the two arguments to be promoted to int64. We pick uint8
-      # here, which will be promoted to uint32 by the binary op.
-      # Note that we prefer unsigned int to signed int when both are equally
-      # precise. For example, for 5, we pick uint8 instead of int8. There is no
-      # reason to prefer one to the other, because for each there is a case
-      # where the behavior diverges from numpy. If we prefer signed int,
-      # consider the case where the first operand is 5 and the second is
-      # 2**64-1. Numpy picks uint64 as the result dtype, but because we choose a
-      # signed type for 5 such as int8, the result type will be float64. On the
-      # other hand, if we prefer unsigned int, consider the case where the first
-      # operand is 2**31-1 and the second is -1. Numpy will pick int32, but
-      # because we choose uint32 for 2*32-1, the result will be int64. The root
-      # of the problem is that `jit` converts `int` to tensors (hence committing
-      # to a dtype) too early, when we don't have enough information about the
-      # jitted function (e.g. which subset of the arguments should be promoted
-      # together using np.result_type). tf.function doesn't have this problem
-      # because it doesn't convert `int` to tensors. jax.jit doesn't have this
-      # problem because it converts `int` to "int tracer" which doesn't commit
-      # to a dtype.
-      # TODO(wangpeng): Revisit this design and see whether we can improve `jit`
-      #   and tf.function.
-      dtype = most_precise_int_dtype(x)
-      if dtype is None and isinstance(x, float):
-        dtype = tf_np.default_float_type()
-      return tf.convert_to_tensor(value=x, dtype=dtype)
-    except (TypeError, ValueError):
-      return x
+    return x
+  try:
+    # We need to convert `int` to the most precise dtype, otherwise the dtype
+    # of the result may be different from numpy's. For example, when a binary
+    # op takes in a Python integer 5 and an array of uint32, numpy will pick
+    # uint32 as 5's dtype, while tf.convert_to_tensor will choose int32 which
+    # will cause the two arguments to be promoted to int64. We pick uint8
+    # here, which will be promoted to uint32 by the binary op.
+    # Note that we prefer unsigned int to signed int when both are equally
+    # precise. For example, for 5, we pick uint8 instead of int8. There is no
+    # reason to prefer one to the other, because for each there is a case
+    # where the behavior diverges from numpy. If we prefer signed int,
+    # consider the case where the first operand is 5 and the second is
+    # 2**64-1. Numpy picks uint64 as the result dtype, but because we choose a
+    # signed type for 5 such as int8, the result type will be float64. On the
+    # other hand, if we prefer unsigned int, consider the case where the first
+    # operand is 2**31-1 and the second is -1. Numpy will pick int32, but
+    # because we choose uint32 for 2*32-1, the result will be int64. The root
+    # of the problem is that `jit` converts `int` to tensors (hence committing
+    # to a dtype) too early, when we don't have enough information about the
+    # jitted function (e.g. which subset of the arguments should be promoted
+    # together using np.result_type). tf.function doesn't have this problem
+    # because it doesn't convert `int` to tensors. jax.jit doesn't have this
+    # problem because it converts `int` to "int tracer" which doesn't commit
+    # to a dtype.
+    # TODO(wangpeng): Revisit this design and see whether we can improve `jit`
+    #   and tf.function.
+    dtype = most_precise_int_dtype(x)
+    if dtype is None and isinstance(x, float):
+      dtype = tf_np.default_float_type()
+    return tf.convert_to_tensor(value=x, dtype=dtype)
+  except (TypeError, ValueError):
+    return x
 
 
 def _canonicalize_jit_arguments(inp):
@@ -107,21 +101,10 @@ def _canonicalize_jit_arguments(inp):
   return tf.nest.map_structure(_canonicalize_jit_arg, inp)
 
 
-def _np_to_tf(inp):
-
-  def f(x):
-    if isinstance(x, tf_np.ndarray):
-      return x.data
-    else:
-      return x
-
-  return tf.nest.map_structure(f, inp)
-
-
 def _tf_to_np(inp):
 
   def f(x):
-    if isinstance(x, (tf.Tensor, tf.IndexedSlices)):
+    if isinstance(x, tf.IndexedSlices):
       return tf_np.asarray(x)
     else:
       return x
@@ -130,7 +113,7 @@ def _tf_to_np(inp):
 
 
 def stop_gradient(x):
-  return _tf_to_np(tf.nest.map_structure(tf.stop_gradient, _np_to_tf(x)))
+  return _tf_to_np(tf.nest.map_structure(tf.stop_gradient, x))
 
 
 def custom_grad(f_vjp, f_original=None):
@@ -162,18 +145,18 @@ def custom_grad(f_vjp, f_original=None):
     np_args = _tf_to_np(tf_args)
     np_kwargs = _tf_to_np(tf_kwargs)
     np_y, np_vjp = f_vjp(*np_args, **np_kwargs)
-    tf_y = _np_to_tf(np_y)
+    tf_y = np_y
 
     def tf_vjp(*flat_tf_dy):
       tf_dy = tf.nest.pack_sequence_as(tf_y, flat_tf_dy)
       np_dy = _tf_to_np(tf_dy)
       np_dx = np_vjp(np_dy)
-      return tf.nest.flatten(_np_to_tf(np_dx))
+      return tf.nest.flatten(np_dx)
 
     return tf_y, tf_vjp
 
   def np_f(*args, **kwargs):
-    return _tf_to_np(tf_f(*_np_to_tf(args), **_np_to_tf(kwargs)))
+    return _tf_to_np(tf_f(*args), **kwargs)
 
   return np_f
 
@@ -200,19 +183,16 @@ def vjp(f, *primals, has_aux=False):
     `y`. `dx` is the cotengents of `x`, having the same structures, shapes and
     dtypes as `x`.
   """
-  tf_primals = _np_to_tf(primals)
   with tf.GradientTape(persistent=True) as tape:
-    tape.watch(tf.nest.flatten(tf_primals))
+    tape.watch(tf.nest.flatten(primals))
     outputs = f(*primals)
     if has_aux:
       np_out, aux = outputs
     else:
       np_out = outputs
-    tf_out = _np_to_tf(np_out)
 
     def _vjp(dy):
-      tf_dy = _np_to_tf(dy)
-      tf_dx = tape.gradient(tf_out, tf_primals, output_gradients=tf_dy)
+      tf_dx = tape.gradient(np_out, primals, output_gradients=dy)
       return _tf_to_np(tf_dx)
 
   if has_aux:
@@ -246,22 +226,21 @@ def grad(f, has_aux=False):
     if not isinstance(np_loss, tf_np.ndarray):
       raise ValueError(
           "The result of the function to take gradient must be an ndarray.")
-    if not np_loss.data.shape.is_compatible_with([]):
+    if not np_loss.shape.is_compatible_with([]):
       raise ValueError(
           "The result of the function to take gradient must be a scalar.")
 
   def _f(params, *args):
     """The gradient function to be returned."""
-    tf_params = _np_to_tf(params)
     with tf.GradientTape() as g:
-      g.watch(tf.nest.flatten(tf_params))
+      g.watch(tf.nest.flatten(params))
       outputs = f(params, *args)
       if has_aux:
         np_loss, aux = outputs
       else:
         np_loss = outputs
       check_loss_shape(np_loss)
-      tf_grads = g.gradient(np_loss.data, tf_params)
+      tf_grads = g.gradient(np_loss, params)
       if has_aux:
         res = (tf_grads, aux)
       else:
@@ -378,7 +357,7 @@ def jit(f,
         np_out = np_out[0]
     else:
       np_out = f(*np_args, **kwargs)
-    return _np_to_tf(np_out)
+    return np_out
 
   def _f(*args, **kwargs):
     args = [
@@ -500,7 +479,7 @@ def _index_update_helper(updater, x, idx, y):
   # TODO(b/164251540): Remove this expensive manual broadcasting once
   #   tf.raw_ops.tensor_strided_slice_update and tf.tensor_scatter_nd_update
   #   support broadcasting.
-  y = tf.broadcast_to(y.data, tf.shape(x[idx].data))
+  y = tf.broadcast_to(y, tf.shape(x[idx]))
   return updater(x, idx, y)
 
 
@@ -610,17 +589,17 @@ def logsumexp(x, axis=None, keepdims=None):
   """
   return tf_np.asarray(
       tf.math.reduce_logsumexp(
-          input_tensor=x.data, axis=axis, keepdims=keepdims))
+          input_tensor=x, axis=axis, keepdims=keepdims))
 
 
 def expit(x):
   """Compute 1 / (1 + exp(-x))."""
-  return tf_np.asarray(tf.math.sigmoid(x.data))
+  return tf_np.asarray(tf.math.sigmoid(x))
 
 
 def erf(x):
   """Computes the Gauss error function of x element-wise."""
-  return tf_np.asarray(tf.math.erf(x.data))
+  return tf_np.asarray(tf.math.erf(x))
 
 
 def _minus(a, b):
@@ -911,8 +890,8 @@ def conv(inp,
   fltr = tf_np.asarray(fltr, dtype)
   return tf_np.asarray(
       tf.nn.convolution(
-          input=inp.data,
-          filters=fltr.data,
+          input=inp,
+          filters=fltr,
           padding=padding,
           strides=window_strides,
           dilations=filter_dilation,
@@ -992,9 +971,9 @@ def sort_key_val(keys, values, dimension=-1):
   """
   keys = tf_np.asarray(keys)
   values = tf_np.asarray(values)
-  rank = keys.data.shape.ndims
+  rank = keys.shape.ndims
   if rank is None:
-    rank = values.data.shape.ndims
+    rank = values.shape.ndims
   if rank is None:
     # We need to know the rank because tf.gather requires batch_dims to be `int`
     raise ValueError("The rank of either keys or values must be known, but "
@@ -1014,11 +993,10 @@ def sort_key_val(keys, values, dimension=-1):
   keys = maybe_swapaxes(keys)
   values = maybe_swapaxes(values)
   idxs = tf_np.argsort(keys)
-  idxs = idxs.data
 
   # Using tf.gather rather than np.take because the former supports batch_dims
   def gather(a):
-    return tf_np.asarray(tf.gather(a.data, idxs, batch_dims=rank - 1))
+    return tf_np.asarray(tf.gather(a, idxs, batch_dims=rank - 1))
 
   keys = gather(keys)
   values = gather(values)
@@ -1062,7 +1040,8 @@ def scan(f, init, xs, length=None, reverse=False):
   """
   init, xs = tf.nest.map_structure(
       lambda x: tf_np.asarray(x) if x is not None else None, (init, xs))
-  init, xs = _np_to_tf((init, xs))
+  if length is not None:
+    length = int(length)
   def get_length(x):
     if x is None:
       return None
@@ -1100,7 +1079,7 @@ def scan(f, init, xs, length=None, reverse=False):
       i_ = i
     xs = tf.nest.map_structure(
         lambda x_ta: x_ta.read(i_) if x_ta is not None else None, xs_ta)
-    carry, ys = _np_to_tf(f(*_tf_to_np((carry, xs))))
+    carry, ys = f(*_tf_to_np((carry, xs)))
     ys_ta = tf.nest.map_structure(
         lambda y_ta, y: (y_ta.write(i_, y) if y is not None else y_ta),
         ys_ta, ys)
@@ -1131,6 +1110,153 @@ def scan(f, init, xs, length=None, reverse=False):
   return _tf_to_np((carry, ys))
 
 
+# named "tf_map" instead of "map" as in JAX to avoid conflict with Python `map`
+def tf_map(f, xs):
+  """Map a function over leading array axes.
+
+  See the docstring of `jax.lax.map`
+  (https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.map.html) for
+  details.
+
+  Args:
+    f: a Python function to apply element-wise over the first axis or axes of
+      `xs`.
+    xs: values over which to map along the leading axis.
+
+  Returns:
+    Mapped values.
+  """
+  def g(unused, x):
+    return unused, f(x)
+  carry = tf.nest.map_structure(lambda _: None, xs)
+  return scan(g, carry, xs)[1]
+
+
+def _get_dynamic_indices(operand, start_indices, slice_sizes):
+  """Calcuates the indices for `tf.gather_nd` from slices.
+
+  Args:
+    operand: a Tensor to slice.
+    start_indices: a vector Tensor of integers, one per dimension. The starts of
+      the slice. The vector can be dynamic.
+    slice_sizes: a list of integers, one per dimension. The sizes of the slice.
+
+  Returns:
+    An index array suitable for `tf.gather_nd` and `tf.scatter_nd`, or `None` if
+    `operand` is a scalar.
+  """
+  rank = len(slice_sizes)
+  operand_rank = tf.rank(operand)
+  tf.debugging.Assert(operand_rank == rank, [operand_rank, rank])
+  starts_rank = tf.rank(start_indices)
+  tf.debugging.Assert(starts_rank == 1, [starts_rank])
+  num_starts = tf.shape(start_indices)[0]
+  tf.debugging.Assert(num_starts == rank, [num_starts, rank])
+  operand_shape = tf.shape(operand)
+  tf.debugging.Assert(tf.reduce_all(slice_sizes <= operand_shape),
+                      [slice_sizes, operand_shape])
+  if rank == 0:
+    return None
+  start_indices = tf.where(
+      start_indices < 0, start_indices + operand_shape, start_indices)
+  idx_list = []
+  for i in range(rank):
+    start = start_indices[i]
+    size = slice_sizes[i]
+    dim = operand_shape[i]
+    start = tf.clip_by_value(start, 0, dim - size)
+    # XLA requires tf.range's `start` to be compile-time constant, so we can't
+    # do tf.range(start, ...).
+    idx = start + tf.range(size)
+    shape = [1] * rank
+    shape[i] = size
+    idx = tf.reshape(idx, shape)
+    idx_list.append(idx)
+  slice_sizes_tensor = tf.convert_to_tensor(slice_sizes)
+  # tf.stack doesn't support broadcasting, so we need to broadcast manually.
+  # TODO(wangpeng): Reduce peak memory by broadcasting one-by-one instead of
+  #   all-together.
+  idx_list = [tf.broadcast_to(x, slice_sizes_tensor) for x in idx_list]
+  return tf.stack(idx_list, axis=-1)
+
+
+def dynamic_slice(operand, start_indices, slice_sizes):
+  """Slicing operation where the indices can be dynamic vlaues.
+
+  See the docstring of `jax.lax.dynamic_slice`
+  (https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.dynamic_slice.html)
+  for details.
+
+  Args:
+    operand: an array to slice.
+    start_indices: a vector of integers, one per dimension. The starts of the
+      slice. The vector can be dynamic.
+    slice_sizes: a list of integers, one per dimension. The sizes of the slice.
+
+  Returns:
+    An array containing the slice, with shape equal to `slice_sizes`.
+  """
+  # This implementation uses tf.gather_nd to implement dynamic_slice, which is
+  # memory inefficient because the size of `indices` given to gather_nd is
+  # large.
+  operand = tf_np.asarray(operand).data
+  start_indices = tf_np.asarray(start_indices, np.int32).data
+  idx = _get_dynamic_indices(operand, start_indices, slice_sizes)
+  if idx is not None:
+    operand = tf.gather_nd(operand, idx)
+  return tf_np.asarray(operand)
+
+
+def dynamic_update_slice(operand, update, start_indices):
+  """Updates a dynamic slice.
+
+  See the docstring of `jax.lax.dynamic_update_slice`
+  (https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.dynamic_update_slice.html)
+  for details.
+
+  Args:
+    operand: an array to slice.
+    update: an array containing the new values to write onto `operand`.
+    start_indices: a vector of integers, one per dimension. The starts of the
+      slice. The vector can be dynamic.
+
+  Returns:
+    The updated version of `operand`.
+  """
+  operand = tf_np.asarray(operand).data
+  update = tf_np.asarray(update).data
+  start_indices = tf_np.asarray(start_indices, np.int32).data
+  if not update.shape.is_fully_defined():
+    raise ValueError("update's shape must be fully defined")
+  slice_sizes = update.shape
+  idx = _get_dynamic_indices(operand, start_indices, slice_sizes)
+  if idx is None:
+    # `np.zeros([])[()] = 1.0` will result in a scalar array of 1.0
+    return tf_np.asarray(update)
+  operand = tf.tensor_scatter_nd_update(operand, idx, update)
+  return tf_np.asarray(operand)
+
+
+def dynamic_slice_in_dim(operand, start_index, slice_size, axis=0):
+  """Convenience wrapper around dynamic_slice applying to one dimension."""
+  operand = tf_np.asarray(operand)
+  start_indices = [0] * operand.ndim
+  slice_sizes = list(operand.shape)
+  axis = int(axis)
+  start_indices[axis] = start_index
+  slice_sizes[axis] = int(slice_size)
+  return dynamic_slice(operand, start_indices, slice_sizes)
+
+
+def dynamic_update_slice_in_dim(operand, update, start_index, axis):
+  """Convenience wrapper around dynamic_update_slice for one dimension."""
+  operand = tf_np.asarray(operand)
+  axis = int(axis)
+  start_indices = [0] * operand.ndim
+  start_indices[axis] = start_index
+  return dynamic_update_slice(operand, update, start_indices)
+
+
 # Use int64 instead of int32 to avoid TF's "int32 problem"
 _RNG_KEY_DTYPE = np.int64
 
@@ -1156,7 +1282,7 @@ def _key2seed(a):
     a = tf.stack(a)
     return a
 
-  return int64_to_int32s(a.data)
+  return int64_to_int32s(a)
 
 
 def _seed2key(a):
@@ -1267,6 +1393,8 @@ def uniform(key,
     An ndarray with shape `shape` and dtype `dtype`. Each value in the ndarray
     is sampled uniformly randomly in range [`minval`, `maxval`).
   """
+  minval = tf.cast(minval, dtype)
+  maxval = tf.cast(maxval, dtype)
   key = tf_np.asarray(key, dtype=_RNG_KEY_DTYPE)
   return tf_np.asarray(
       tf.random.stateless_uniform(
@@ -1502,7 +1630,7 @@ def _psum(tensor, axis_name=None):
       tensor = tf.cast(tensor, tf.float64)
   else:
     tensor = tf.raw_ops.CollectiveReduce(
-        input=tensor.data,
+        input=tensor,
         group_size=len(devices),
         group_key=_GROUP_KEY,
         instance_key=_get_instance_key(),
@@ -1541,7 +1669,7 @@ def pmean(tensor, axis_name=None):
     raise ValueError("pmean for TPU is not supported yet.")
   else:
     return tf.raw_ops.CollectiveReduce(
-        input=tensor.data,
+        input=tensor,
         group_size=len(devices),
         group_key=_GROUP_KEY,
         instance_key=_get_instance_key(),
@@ -1574,7 +1702,7 @@ def _get_pmap_impl(f, devices, has_tpu):
     """A wrapper for `f` that takes/returns tensors."""
     np_args = _tf_to_np(tf_args)
     np_out = f(*np_args)
-    return _np_to_tf(np_out)
+    return np_out
 
   if has_tpu:
 
@@ -1647,7 +1775,6 @@ def pmap(f, axis_name=None, devices=None):
                        "yet.")
     # TODO(wangpeng): Maybe we should use `asarray` to convert everything
     # to ndarray first.
-    args = _np_to_tf(args)
 
     flattened_input_args = tf.nest.flatten(args)
     flattened_per_device_args = [[] for _ in devices]
