@@ -1,5 +1,5 @@
 # coding=utf-8
-# Copyright 2020 The Trax Authors.
+# Copyright 2021 The Trax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,11 +14,6 @@
 # limitations under the License.
 
 """Tests for tf numpy mathematical methods."""
-
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import functools
 import itertools
 
@@ -73,16 +68,12 @@ def uniform(rng, shape, dtype):
   return tf_np.asarray(rng.uniform(shape=shape, dtype=dtype, minval=minval))
 
 
-def to_tf(a):
-  return tf.nest.map_structure(lambda x: x.data, a)
-
-
 def to_np(a):
   return tf.nest.map_structure(tf_np.asarray, a)
 
 
 def to_tf_fn(f):
-  return lambda *args: to_tf(f(*to_np(args)))
+  return lambda *args: f(*to_np(args))
 
 
 def scan_reference(f, init, xs):
@@ -144,8 +135,8 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
     expected_dx = (dy * scale1 * x[0], dy * scale2 * x[1])
     y, vjp = extensions.vjp(f, *x)
     dx = vjp(dy)
-    self.assertAllClose(to_tf(expected_y), to_tf(y))
-    self.assertAllClose(to_tf(expected_dx), to_tf(dx))
+    self.assertAllClose(expected_y, y)
+    self.assertAllClose(expected_dx, dx)
 
   @parameterized.named_parameters([
       (  # pylint: disable=g-complex-comprehension
@@ -204,7 +195,7 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
     rng = tf.random.Generator.from_seed(1234)
     x, dy_list = tf.nest.map_structure(lambda shape: uniform(rng, shape, dtype),
                                        [x_shape, [y_shape] * 2])
-    tf_x = to_tf(x)
+    tf_x = x
     outputs = extensions.vjp(f, *x, has_aux=has_aux)
     if has_aux:
       y, vjp, aux = outputs
@@ -215,14 +206,14 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
       outputs = f(*x)
       if has_aux:
         expected_y, expected_aux = outputs
-        self.assertAllClose(to_tf(expected_aux), to_tf(aux))
+        self.assertAllClose(expected_aux, aux)
       else:
         expected_y = outputs
-    self.assertAllClose(to_tf(expected_y), to_tf(y))
+    self.assertAllClose(expected_y, y)
     for dy in dy_list:
       expected_dx = tape.gradient(
-          to_tf(expected_y), tf_x, output_gradients=to_tf(dy))
-      self.assertAllClose(expected_dx, to_tf(vjp(dy)))
+          expected_y, tf_x, output_gradients=dy)
+      self.assertAllClose(expected_dx, vjp(dy))
 
   def testGrad(self):
 
@@ -233,9 +224,9 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
 
     def compare(a, b):
       with tf.GradientTape() as tape:
-        tape.watch(a.data)
+        tape.watch(a)
         r = f(a, b)
-      expected = tape.gradient(r.data, a.data)
+      expected = tape.gradient(r, a)
       self.assertAllEqual(expected, g(a, b))
 
     shape = [10]
@@ -506,7 +497,7 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
                                                  feature_group_count,
                                                  batch_group_count)
 
-    self.assertAllEqual(tf_conv, tf_np.asarray(jax_conv))
+    self.assertAllClose(tf_conv, tf_np.asarray(jax_conv))
 
   def testConv(self):
     y = extensions.conv(
@@ -684,6 +675,22 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
         functools.partial(extensions.scan, f), static_argnums=(2,))(*inputs)
     self.assertAllEqual(expected_outputs, outputs)
 
+  def testMap(self):
+    shape = [2, 3]
+    dtype = tf_np.int32
+    xs1 = tf_np.zeros(shape, dtype)
+    xs2 = tf_np.ones(shape, dtype)
+    ys_expected = [xs2 + 10, xs1 + 20]
+    def f(x):
+      self.assertIsInstance(x, tuple)
+      for a in x:
+        self.assertEqual(a.shape, shape[1:])
+      x1, x2 = x
+      return [x2 + 10, x1 + 20]
+    ys = extensions.tf_map(f, (xs1, xs2))
+    self.assertIsInstance(ys, list)
+    self.assertAllClose(ys, ys_expected)
+
   def testPrng(self):
     self.assertAllEqual(tf_np.asarray(123, np.int64), extensions.prng(123))
 
@@ -712,8 +719,7 @@ class ExtensionsTest(tf.test.TestCase, parameterized.TestCase):
   def testBernoulliWrongShape(self):
     mean = [0.1, 0.2]
     shape = [3]
-    with self.assertRaisesWithPredicateMatch(tf.errors.InvalidArgumentError,
-                                             r"Incompatible shapes"):
+    with self.assertRaisesIncompatibleShapesError():
       extensions.bernoulli(123, mean, shape)
 
   def testDatasetAsNumpy(self):
